# -*- coding: utf-8 -*-
import json
import logging
import os
import tempfile
from pathlib import Path

import requests
from apistar.exceptions import ErrorResponse
from teklia_toolbox.requests import download_file

from arkindex_tasks import default_client
from arkindex_tasks.enums import ProcessMode

logger = logging.getLogger(__name__)

WORKER_RUN_ID = os.environ.get("ARKINDEX_WORKER_RUN_ID")

_working_dir = None


def get_working_dir():
    global _working_dir
    if _working_dir:
        return _working_dir

    if "PONOS_DATA" in os.environ:
        _working_dir = Path(os.environ["PONOS_DATA"]) / "current"
    else:
        _working_dir = Path(tempfile.mkdtemp())
    return _working_dir


def _get_id(element):
    if isinstance(element, dict) and element.get("id"):
        return element["id"]
    return str(element)


def dump_json(elements, path=None, filename="elements.json"):
    file_path = path and Path(path, filename)
    if not file_path:
        file_path = get_working_dir() / filename
    assert len(elements), "No elements could be written to {}".format(file_path)

    with file_path.open("w") as f:
        json.dump(elements, f, indent=4)


def split_chunks(items, n):
    """
    Yield n number of elements from a given list with a balanced distribution
    https://stackoverflow.com/questions/24483182/python-split-list-into-n-chunks#answer-54802737
    """
    for i in range(0, n):
        yield items[i::n]


class ProcessTask(object):
    """
    Base class for tasks based on Processes.
    """

    def __init__(self, process_id):
        self.process_id = process_id
        self.process = None
        self.folder = None

    def load_process(self):
        self.process = default_client.request(
            "RetrieveProcess", id=str(self.process_id)
        )

    @property
    def mode(self):
        if self.process is None:
            return
        return ProcessMode(self.process["mode"])

    def download_files(self):
        temp_path = tempfile.mkdtemp()
        files = []
        for datafile_id in self.process["files"]:
            df = default_client.request("RetrieveDataFile", id=datafile_id)
            path = Path(temp_path) / datafile_id

            logger.info("Downloading file {} ({})".format(df["name"], df["id"]))
            try:
                download_file(df["s3_url"], path)
            except requests.RequestException as e:
                logger.error(
                    "Failed downloading file {} ({}): {}".format(
                        df["name"], df["id"], str(e)
                    )
                )
                continue

            df["local_path"] = path
            files.append(df)

        return files

    def delete_files(self, files):
        """
        Delete DataFiles once they are imported.
        Deletion errors will only show warnings instead of making the import fail.
        """
        for datafile in files:
            try:
                default_client.request("DestroyDataFile", id=datafile["id"])
            except ErrorResponse as e:
                logger.warning(f'Could not delete file {datafile["name"]}: {e}')
