# -*- coding: utf-8 -*-
import hashlib
import logging
import os
import shutil
import sys
import tarfile
from datetime import datetime
from tempfile import NamedTemporaryFile, TemporaryFile
from zipfile import BadZipFile, ZipFile

import requests
from apistar.exceptions import ErrorResponse
from natsort import natsorted
from PIL import Image
from teklia_toolbox.requests import should_verify_cert
from tenacity import retry, retry_if_exception_type, stop_after_attempt, wait_fixed
from zstandard import ZstdDecompressor

import magic
from arkindex_tasks import default_client
from arkindex_tasks.base import WORKER_RUN_ID, ProcessTask, dump_json
from arkindex_tasks.import_files.iiif import IIIFParser
from arkindex_tasks.import_files.image import check_image
from arkindex_tasks.import_files.pdf import extract_pdf_images, upload_pdf_text
from arkindex_tasks.import_files.transkribus import TranskribusImporter

logger = logging.getLogger(__name__)

# application/zip is the standard, but all browsers on Windows
# upload files with application/x-zip-compressed instead.
ZIP_MIME_TYPES = {
    "application/zip",
    "application/x-zip-compressed",
}

IIIF_MIME_TYPES = {
    "application/json",
    "application/ld+json",
}

TAR_MIME_TYPES = {
    # .tar
    "application/x-tar",
    # .tar.gz, but browsers will usually only detect the .gz
    "application/x-gtar",
    # .gz
    "application/gzip",
    # .bz2
    "application/x-bzip",
    "application/x-bzip2",
    # .lzma
    "application/x-lzma",
    # .xz
    "application/x-xz",
}

ZSTANDARD_MIME_TYPES = {
    "application/zstd",
}

ARCHIVE_MIME_TYPES = ZIP_MIME_TYPES | TAR_MIME_TYPES | ZSTANDARD_MIME_TYPES

# Maximum nesting of archives (file.zip.zip.zip).
# Helps protect against ZIP quines like <https://alf.nu/ZipQuine>,
# but not ZIP bombs like <https://www.bamsoftware.com/hacks/zipbomb/>,
# as they do not need to be recursive to be radioactive.
MAX_ARCHIVE_DEPTH = 3


class FileImport(ProcessTask):
    """
    A generic file import: takes a Process and creates elements in a folder from its files.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.elements = []
        self.archive_depth = 0

    def create_top_level_folder(self, datafiles):
        """
        When a process runs without a parent element, files that have a single image
        (images and single-page PDFs) would get created without a parent folder, so they
        would be invisible to the user.
        Here, we create a top-level folder for any import with images only.
        `save_elements` includes will do it for PDFs, and `import_archive` for archives.
        This receives the datafiles as an argument to deduce the folder's name.
        """
        if self.process.get("element") or not all(
            datafile["content_type"].startswith("image/") for datafile in datafiles
        ):
            return

        if len(datafiles) == 1:
            parent_name = datafiles[0]["name"]
        else:
            parent_name = (
                f'Import images {datetime.now().isoformat(sep=" ", timespec="minutes")}'
            )

        self.process["element"] = default_client.request(
            "CreateElement",
            body={
                "corpus": self.process["corpus"],
                "type": self.process["folder_type"],
                "name": parent_name,
                "worker_run_id": WORKER_RUN_ID,
            },
        )

    def get_images(self, datafile):
        """
        From a given DataFile, returns a list of paths to images
        compatible with an IIIF server and ML operations
        """
        try:
            if datafile["content_type"].startswith("image/"):
                return check_image(datafile)
            if datafile["content_type"] == "application/pdf":
                return extract_pdf_images(datafile)
        except Exception as e:
            logger.warning(f"Could not fetch images from {datafile['name']}: {e}")
            return []

        raise NotImplementedError

    def hash_image(self, path):
        md5 = hashlib.md5()
        with open(path, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                md5.update(chunk)
        return md5.hexdigest()

    def create_image(self, path):
        logger.debug("Creating image {}".format(path))
        hash = self.hash_image(path)

        try:
            image = default_client.request("CreateImage", body={"hash": hash})
        except ErrorResponse as e:
            if e.status_code == 400 and "id" in e.content:
                image = default_client.request("RetrieveImage", id=e.content["id"])
            else:
                raise
        else:
            logger.info("Uploading image {}".format(image["id"]))
            self.upload_image(path, image["s3_put_url"])

        logger.debug("Validating image {}".format(image["id"]))
        default_client.request(
            "PartialUpdateImage", id=image["id"], body={"status": "checked"}
        )

        return image["id"]

    @retry(
        reraise=True,
        retry=retry_if_exception_type(requests.RequestException),
        stop=stop_after_attempt(3),
        wait=wait_fixed(5),
    )
    def upload_image(self, path, url):
        with open(path, "rb") as f:
            resp = requests.put(url, data=f, verify=should_verify_cert(url))
        resp.raise_for_status()

    @retry(
        reraise=True,
        retry=retry_if_exception_type(ErrorResponse),
        stop=stop_after_attempt(3),
        wait=wait_fixed(5),
    )
    def create_element(self, image_path, name, parent_id):
        assert parent_id is not None, "Cannot create an image as a top-level element."
        return default_client.request(
            "CreateElement",
            body={
                "corpus": self.process["corpus"],
                "type": self.process["element_type"],
                "name": name,
                "image": self.create_image(image_path),
                "parent": parent_id,
                "worker_run_id": WORKER_RUN_ID,
            },
        )

    def save_elements(self, datafile, image_paths, parent_id):
        # When there is a single image, create the element immediately without a parent
        # If the import runs without a parent folder (on a corpus directly), we will still create
        # a parent folder to avoid elements that are invisible in the API.
        if len(image_paths) == 1 and parent_id:
            return [
                self.create_element(
                    image_paths[0],
                    datafile["name"],
                    parent_id,
                )
            ]

        # Create a parent folder for a file with multiple images
        parent = default_client.request(
            "CreateElement",
            body={
                "parent": parent_id,
                "corpus": self.process["corpus"],
                "type": self.process["folder_type"],
                "name": datafile["name"],
                "worker_run_id": WORKER_RUN_ID,
            },
        )
        parent_id = parent["id"]

        elements = [parent]
        elements.extend(
            [
                self.create_element(image_path, str(i), parent_id)
                for i, image_path in enumerate(image_paths, start=1)
            ]
        )
        return elements

    def is_transkribus_export(self, datafile) -> bool:
        try:
            # Opening the file with ZipFile will raise a BadZipFile exception
            # when the file is not a valid ZIP file.
            with ZipFile(datafile["local_path"]) as archive:
                # To detect that an archive is a Transkribus export, we look for a `mets.xml` file,
                # which should be present in every export, under the subdirectories for each document.
                return any(
                    filename.endswith("/mets.xml") for filename in archive.namelist()
                )
        except BadZipFile:
            logger.error(
                f"File {datafile['name']} does not appear to be a valid ZIP archive."
            )
            return False

    def import_transkribus_archive(self, datafile, parent_id) -> list:
        logger.info(f'Starting Transkribus import for {datafile["name"]}')

        return TranskribusImporter(
            datafile["local_path"],
            corpus=self.process["corpus"],
            parent_id=parent_id,
            folder_type=self.process["folder_type"],
            element_type=self.process["element_type"],
        ).run()

    def get_archive_datafiles(self, datafile):
        """
        Generator that returns DataFile dicts for each file within an archive DataFile.
        The `local_path` is only guaranteed to exist until the next `next()` call.
        """
        if datafile["content_type"] in ZIP_MIME_TYPES:
            with ZipFile(datafile["local_path"]) as archive:
                for fileinfo in archive.infolist():
                    with archive.open(
                        fileinfo
                    ) as source, NamedTemporaryFile() as extracted_file:
                        shutil.copyfileobj(source, extracted_file)
                        # Very small files may not be fully written to disk, which could cause
                        # the MIME type detection to find an empty file
                        extracted_file.flush()

                        yield {
                            "id": datafile["id"],
                            "name": fileinfo.filename,
                            "size": fileinfo.file_size,
                            "status": "checked",
                            # The file import relies on MIME types assigned to DataFiles,
                            # but we do not have them from within archives, so guess them with libmagic
                            "content_type": magic.from_file(
                                extracted_file.name, mime=True
                            ),
                            "local_path": extracted_file.name,
                        }

        elif datafile["content_type"] in TAR_MIME_TYPES | ZSTANDARD_MIME_TYPES:
            if datafile["content_type"] in ZSTANDARD_MIME_TYPES:
                # TarFile does not support .tar.zst, so we decompress ourselves
                fileobj = TemporaryFile()
                with open(datafile["local_path"], "rb") as compressed_stream:
                    ZstdDecompressor().copy_stream(compressed_stream, fileobj)
                fileobj.seek(0)
            else:
                # Just open the file and let Tarfile deal with other compression methods
                fileobj = open(datafile["local_path"], "rb")

            # We have to include `with fileobj` because TarFile() will not close the original file object by itself
            with fileobj, tarfile.open(fileobj=fileobj) as archive:
                # Block all of the tar features that could cause security issues
                # See <https://docs.python.org/3/library/tarfile.html#tarfile-extraction-filter>
                archive.extraction_filter = tarfile.data_filter

                for fileinfo in archive.getmembers():
                    # Skip directories, symbolic and hard links, character/block devices and FIFOs
                    if not fileinfo.isfile():
                        continue

                    with archive.extractfile(
                        fileinfo
                    ) as source, NamedTemporaryFile() as extracted_file:
                        shutil.copyfileobj(source, extracted_file)

                        yield {
                            "id": datafile["id"],
                            "name": fileinfo.name,
                            "size": fileinfo.size,
                            "status": "checked",
                            # The file import relies on MIME types assigned to DataFiles,
                            # but we do not have them from within archives, so guess them with libmagic
                            "content_type": magic.from_file(
                                extracted_file.name, mime=True
                            ),
                            "local_path": extracted_file.name,
                        }

        else:
            raise NotImplementedError

    def import_archive(self, datafile, parent_id) -> None:
        """
        Import an archive, as if we were importing all of its files as separate DataFiles in a subfolder.
        Subfolders within the archive are not supported; the datafiles will just have slashes in their names.

        When any file in the archive has not been successfully imported, an exception is raised,
        but only after having tried to import as much of the archive as possible.
        """
        assert (
            self.archive_depth < MAX_ARCHIVE_DEPTH
        ), "Maximum archive decompression depth reached"
        self.archive_depth += 1

        logger.info(f'Creating folder for archive {datafile["name"]}')
        archive_folder = default_client.request(
            "CreateElement",
            body={
                "corpus": self.process["corpus"],
                "type": self.process["folder_type"],
                "parent": parent_id,
                "name": datafile["name"],
                "worker_run_id": WORKER_RUN_ID,
            },
        )

        logger.info(f"Importing files from archive {datafile['name']}")

        failed = False
        for subdatafile in self.get_archive_datafiles(datafile):
            try:
                self.import_datafile(subdatafile, archive_folder["id"])
            except Exception as e:
                logger.error(
                    f'Could not import file {subdatafile["name"]} from archive {datafile["name"]}: {e.__class__.__name__}: {str(e)}'
                )
                failed = True

        self.archive_depth -= 1

        assert (
            not failed
        ), f'Not all files from archive {datafile["name"]} have been imported.'

    def import_iiif(self, datafile, parent_id) -> None:
        try:
            with datafile["local_path"].open("rb") as f:
                parser = IIIFParser.load(
                    f,
                    folder_type=self.process["folder_type"],
                    element_type=self.process["element_type"],
                    corpus_id=self.process["corpus"],
                    parent_id=parent_id,
                )
                parser.run()
            self.elements.extend(parser.elements)
        except ErrorResponse as e:
            logger.error(
                "IIIF import failed for file {!s} ({!s}): {!s} {!r}".format(
                    datafile["name"], datafile["id"], e.status_code, e.content
                ),
                exc_info=False,
            )
        except Exception as e:
            logger.error(
                "IIIF import failed for file {!s} ({!s}): {!r}".format(
                    datafile["name"], datafile["id"], e
                ),
                exc_info=False,
            )

    def import_datafile(self, datafile, parent_id) -> None:
        """
        Import a single DataFile into a parent folder, and add the created elements to `self.elements`.

        When a file has not been successfully imported in its entirety, exceptions may be raised.
        """
        if datafile["content_type"] in ARCHIVE_MIME_TYPES:
            if datafile[
                "content_type"
            ] in ZIP_MIME_TYPES and self.is_transkribus_export(datafile):
                self.import_transkribus_archive(datafile, parent_id)
            else:
                self.import_archive(datafile, parent_id)
            return

        # The ld+json format support parameters separated by a semicolon,
        # like application/ld+json;profile="http://iiif.io/api/presentation/2/context.json";charset=utf-8
        if datafile["content_type"].split(";")[0] in IIIF_MIME_TYPES:
            self.import_iiif(datafile, parent_id)
            return

        logger.info(f"Fetching images for {datafile['name']}")
        images = self.get_images(datafile)
        if not images:
            return

        logger.info(f"Creating {len(images)} elements")
        new_elements = self.save_elements(datafile, images, parent_id)
        self.elements.extend(new_elements)

        # Add text extraction on PDF files
        if datafile["content_type"] == "application/pdf":
            logger.info(f"Extracting text from PDF file {datafile['name']}")

            # We only need the ordered list of pages created
            # A folder may be created for PDFs with multiple pages and included in the list
            pages = [
                elt
                for elt in new_elements
                if elt["type"] == self.process["element_type"]
            ]

            upload_pdf_text(datafile["local_path"], pages)

    def run(self):
        assert WORKER_RUN_ID, "A WorkerRun ID is required"

        max_image_pixels = os.environ.get("ARKINDEX_MAX_IMAGE_PIXELS", "")
        if max_image_pixels:
            max_image_pixels = int(max_image_pixels)
            # Override Pillow limit for detecting decompression bombs, disabled if set to 0
            if max_image_pixels == 0:
                logger.warning(
                    "Pillow Image size limit is completely disabled, make sure you trust the image source."
                )
                Image.MAX_IMAGE_PIXELS = None
            else:
                Image.MAX_IMAGE_PIXELS = max_image_pixels

        logger.info("Running import {!s}".format(self.process_id))
        self.load_process()

        logger.info("Downloading files...")
        datafiles = self.download_files()
        if len(datafiles) > 0:
            logger.info("Downloaded {} files".format(len(datafiles)))
        else:
            logger.error("No files downloaded")
            sys.exit(1)

        # Sort files by name as a human would expect: a1.jpg < a10.jpg
        datafiles = natsorted(datafiles, key=lambda datafile: datafile["name"])

        self.create_top_level_folder(datafiles)

        parent_id = None
        if self.process["element"]:
            parent_id = self.process["element"]["id"]
            logger.info(
                f'Using folder {self.process["element"]["id"]} "{self.process["element"]["name"]}"'
            )
            self.elements.append(
                {
                    "type": self.process["element"]["type"],
                    "id": self.process["element"]["id"],
                }
            )

        successful_datafiles = []
        for df in datafiles:
            try:
                self.import_datafile(df, parent_id)
            except Exception as e:
                logger.error(
                    f'Could not import file {df["name"]}: {e.__class__.__name__}: {str(e)}'
                )
                continue

            successful_datafiles.append(df)

        if len(self.elements) < 1:
            logger.error("No elements have been imported")
            sys.exit(1)

        logger.info(f"Imported {len(self.elements) - 1} elements")

        # The file import might create any amount of folders, but the first one
        # is always the topmost one. We look for it in the elements list then
        # update the Process's element so that the frontend can display
        # a "View element" button.
        created_folder = next(
            (
                element
                for element in self.elements
                if element["type"] == self.process["folder_type"]
            ),
            None,
        )
        if created_folder:
            default_client.request(
                "PartialUpdateProcess",
                id=self.process["id"],
                body={"element_id": created_folder["id"]},
            )

        logger.info("Deleting files…")
        self.delete_files(successful_datafiles)

        dump_json(self.elements)
