# -*- coding: utf-8 -*-
import logging
from io import BufferedIOBase, BytesIO
from pathlib import Path
from urllib.parse import urljoin, urlparse

import ijson
import requests
from apistar.exceptions import ErrorResponse
from teklia_toolbox.requests import retried_request

from arkindex_tasks import USER_AGENT
from arkindex_tasks.base import WORKER_RUN_ID
from arkindex_tasks.enums import MetaType
from arkindex_tasks.import_files.utils import retried_iiif_get

logger = logging.getLogger(__name__)

# Timeout for the info.json request on images
IIIF_INFO_TIMEOUT = 15


def to_primitive(data, many=False):
    """
    Ensures a JSON-LD value is of a normal primitive type.
    Values can come in three forms: normal primitives, value objects and arrays of value objects.
    A JSON-LD value object has an @value key, and optionally @type, @language, or @index.
    See https://json-ld.org/spec/latest/json-ld/#value-objects

    In the case of multiple values objects, this picks the first object with "@language": "en" or starting with "en-",
    as Arkindex is in English and not internationalized, or just picks the first found value.

    If many is set to True, this always returns a list;
    when there are multiple values and some do not have an associated language, this returns all of the values.
    """
    if isinstance(data, list):
        if len(data) <= 0:
            raise ValueError("Array of JSON-LD Value Objects should not be empty")
        all_dicts = all(isinstance(item, dict) for item in data)
        if not many and not all_dicts:
            raise ValueError("Array should only hold JSON-LD Value Objects")

        if many and (not all_dicts or not all(item.get("@language") for item in data)):
            # No languages, or some items are not JSON-LD objects; apply to_primitive
            return list(filter(None, map(to_primitive, data)))

        data = next(
            (
                item
                for item in data
                if item.get("@language") == "en"
                or item.get("@language", "").startswith("en-")
            ),
            data[0],
        )

    if isinstance(data, dict):
        if "@value" not in data:
            raise ValueError(
                "Expected primitive or JSON-LD Value Object, got an object without a @value property"
            )
        if many:
            return [data["@value"]]
        return data["@value"]

    if many:
        return [data]
    return data


def trim_warn(text, name="Value", max_length=250):
    if len(text) <= max_length:
        return text
    logger.warning(
        "{} is longer than the {} character limit and will be trimmed".format(
            name, max_length
        )
    )
    return text[:max_length]


class IIIFParser(object):
    """
    Base IIIF parser
    """

    def __init__(
        self,
        stream,
        *,
        corpus_id=None,
        parent_id=None,
        folder_type="folder",
        element_type="page",
        folder_name=None,
        **kwargs,
    ):
        """
        :param stream BufferedIOBase: File-like object in binary mode for the JSON document.
        :param corpus_id: A corpus UUID to create elements in.
        :type corpus_id: UUID or str
        :param folder_type str: Slug of the folder type to use.
        :param element_type str: Slug of the element type to use.
        :param parent_id: A parent folder UUID for nested imports.
        :type parent_id: UUID or str or None
        :param folder_name: Optionally override the folder name.
        """
        assert isinstance(
            stream, BufferedIOBase
        ), "Stream should be a file-like object in binary mode"
        self.stream = stream
        self.corpus_id = str(corpus_id)
        self.parent_id = str(parent_id) if parent_id else None
        self.folder_type = folder_type
        self.element_type = element_type
        self.folder_name = folder_name
        self.folder = None
        self.elements = []

    def close(self):
        """
        Closes the stream if it is open.
        """
        if self.stream and not self.stream.closed:
            self.stream.close()

    def check_version(self):
        """
        Raise an error if the manifest is IIIF v3
        """
        context = next(
            (
                value
                for prefix, event, value in ijson.parse(self.stream)
                if prefix == "@context"
            ),
            None,
        )
        if context and "http://iiif.io/api/presentation/3/context.json" in context:
            raise ValueError(
                "The import functionality does not support IIIF Presentation API version 3"
            )
        self.stream.seek(0)

    def get_type(self):
        """
        Get a parser class for the given IIIF object.
        Raises ValueError if the object has no @type or if the type is not supported.
        """
        jsonld_type = next(
            (
                value
                for prefix, event, value in ijson.parse(self.stream)
                if (prefix, event) == ("@type", "string")
            ),
            None,
        )

        if jsonld_type is None:
            raise ValueError("Missing @type property in JSON data")

        self.stream.seek(0)
        try:
            return PARSERS[jsonld_type]
        except KeyError:
            raise ValueError('Unsupported @type "{}"'.format(jsonld_type))

    @staticmethod
    def load(stream, *args, **kwargs):
        p = IIIFParser(stream)
        p.check_version()
        cls = p.get_type()
        return cls(stream, *args, **kwargs)

    def run(self):
        raise NotImplementedError

    def create_element(self, data, iiif_id=None):
        """
        Create an element and add it for elements.json
        """
        element = retried_request("CreateElement", body=data)
        self.elements.append({"id": element["id"], "type": element["type"]})

        if iiif_id:
            retried_request(
                "CreateMetaData",
                id=element["id"],
                body={
                    "name": "IIIF ID",
                    "type": MetaType.Reference.value,
                    "value": iiif_id,
                    "worker_run_id": WORKER_RUN_ID,
                },
            )

        return element

    def create_folder(self):
        # TODO: Handle existing folders
        if not self.folder_name:
            # No overridden label, try to use the document's label
            self.stream.seek(0)
            self.folder_name = to_primitive(
                next(ijson.items(self.stream, "label"), None)
            )

        # The spec says a label is required, but the spec is rarely followed
        # Fallback to the filename when available
        if not self.folder_name and hasattr(self.stream, "name"):
            self.folder_name = Path(self.stream.name).stem
        assert self.folder_name, "Could not determine a folder name"

        # Get the @id to save as a metadata if it exists
        self.stream.seek(0)
        iiif_id = to_primitive(next(ijson.items(self.stream, "@id"), None))

        body = {
            "corpus": self.corpus_id,
            "type": self.folder_type,
            "name": trim_warn(self.folder_name, name="Manifest label"),
            "worker_run_id": WORKER_RUN_ID,
        }
        if self.parent_id:
            body["parent"] = self.parent_id

        self.folder = self.create_element(body, iiif_id)
        logger.info(
            "Created folder {} ({})".format(self.folder["name"], self.folder["id"])
        )

    def create_metadata(self, element, metadatas):
        logger.info(
            "Parsing metadata for element {} ({})".format(
                element["name"], element["id"]
            )
        )
        for i, metadata in enumerate(metadatas):
            label = to_primitive(metadata.get("label"))
            values = to_primitive(metadata.get("value"), many=True)
            if not label:
                logger.warning("Metadata {} does not have a label".format(i))
                continue
            if not any(values):
                logger.warning("Metadata {!r} does not have a value".format(label))
                continue

            for value in values:
                meta_type = MetaType.Text
                if urlparse(value).scheme in ("http", "https"):
                    meta_type = MetaType.URL

                try:
                    retried_request(
                        "CreateMetaData",
                        id=element["id"],
                        body={
                            "name": trim_warn(
                                label, name="Label for metadata {}".format(i)
                            ),
                            "type": meta_type.value,
                            "value": value,
                            "worker_run_id": WORKER_RUN_ID,
                        },
                    )
                    logger.info(
                        "Created metadata {!r} on element {} ({})".format(
                            label, element["name"], element["id"]
                        )
                    )
                except ErrorResponse as e:
                    logger.warning(
                        "Failed creating metadata {!r}: {!s}".format(label, e),
                        exc_info=False,
                    )

    def create_folder_metadata(self):
        self.stream.seek(0)
        self.create_metadata(self.folder, ijson.items(self.stream, "metadata.item"))


class ManifestParser(IIIFParser):
    def get_image(self, url):
        """
        Get or create an Arkindex image from an IIIF image URL.
        Returns None if the image could not be created.
        """

        if not url.endswith("/"):
            url += "/"
        info_url = urljoin(url, "info.json")
        try:
            payload = self.fetch_image_information(info_url)
        except Exception as e:
            logger.warning(
                f"Could not fetch image information at {info_url}: {e}", exc_info=False
            )
            return

        try:
            image = retried_request("CreateIIIFInformation", body=payload)
        except ErrorResponse as e:
            if (
                e.status_code != 400
                or "status" not in e.content
                or "id" not in e.content
            ):
                raise
            logger.warning(
                f'Image check failed for {url}: {e.content["status"]}', exc_info=False
            )
            return retried_request("RetrieveImage", id=e.content["id"])

        return image

    def fetch_image_information(self, info_url):
        resp = retried_iiif_get(
            info_url,
            headers={"User-Agent": USER_AGENT},
            timeout=IIIF_INFO_TIMEOUT,
            allow_redirects=True,
        )
        return resp.json()

    def parse_canvas(self, canvas, default_name=None):
        name = canvas.get("label", default_name)
        if not name:
            raise ValueError("Canvas has no name")

        images = canvas.get("images")
        if not images:
            raise ValueError("Canvas {} has no image".format(name))
        if len(images) > 1:
            logger.warning(
                "Canvas {} has multiple images; will only use the first one".format(
                    name
                )
            )

        resource = images[0].get("resource")
        if resource is None:
            raise ValueError("Canvas {} image has no resource".format(name))

        if "service" not in resource:
            raise ValueError(
                "Image resource {} has no IIIF image information service".format(
                    resource.get("@id")
                )
            )
        if "@id" not in resource["service"]:
            raise ValueError(
                "Image service on resource {} has no @id".format(resource.get("@id"))
            )
        service_id = resource["service"]["@id"]

        image = self.get_image(service_id)

        # TODO: URL parsing to allow elements on portions of an image
        element = self.create_element(
            {
                "corpus": self.corpus_id,
                "parent": self.folder["id"],
                "type": self.element_type,
                "image": image["id"] if image else None,
                "name": trim_warn(name, name="Canvas label"),
                "worker_run_id": WORKER_RUN_ID,
            },
            iiif_id=canvas.get("@id"),
        )
        logger.info("Created element {} ({})".format(element["name"], element["id"]))

        if "metadata" in canvas:
            assert isinstance(canvas["metadata"], list)
            self.create_metadata(element, canvas["metadata"])

    def parse_canvases(self):
        logger.info("Parsing canvases")
        self.stream.seek(0)
        for i, canvas in enumerate(
            ijson.items(self.stream, "sequences.item.canvases.item")
        ):
            try:
                self.parse_canvas(canvas, default_name=str(i))
            except Exception as e:
                logger.warning(
                    "Failed parsing canvas #{} ({}): {!s}".format(
                        i, canvas.get("@id"), e
                    ),
                    exc_info=False,
                )
                continue

    def run(self):
        assert WORKER_RUN_ID, "A WorkerRun ID is required"
        try:
            self.create_folder()
            self.create_folder_metadata()
            self.parse_canvases()
        finally:
            self.close()


class CollectionParser(IIIFParser):
    def __init__(self, stream, *args, first_level=True, **kwargs):
        """
        :param first_level bool: When True, will not create a folder to hold all of the collection's members
           (when importing a collection as the first level of a corpus)
        """
        super().__init__(stream, *args, **kwargs)
        self.parser_args = args
        self.parser_kwargs = kwargs
        self.first_level = first_level

    def create_folder(self):
        if self.first_level:
            return
        super().create_folder()
        self.parser_kwargs["parent_id"] = self.folder["id"]

    def _run_from(self, prop):
        logger.info("Reading collection members in {!r}".format(prop))
        self.stream.seek(0)
        i = -1
        for i, data in enumerate(ijson.items(self.stream, "{}.item".format(prop))):
            if "@id" not in data:
                logger.warning("No @id defined for item {}".format(i))
                continue

            try:
                resp = requests.get(data["@id"], timeout=30)
                resp.raise_for_status()
            except requests.RequestException as e:
                logger.warning(
                    "Failed fetching manifest/collection at {}: {!s}".format(
                        data["@id"], e
                    ),
                    exc_info=False,
                )
                continue

            # Override the folder name with the label from the collection when available
            try:
                parser = IIIFParser.load(
                    BytesIO(resp.content),
                    *self.parser_args,
                    **self.parser_kwargs,
                    first_level=False,
                    folder_name=to_primitive(data.get("label")),
                )
            except ValueError as e:
                logger.warning("{!s} for {}".format(e, data["@id"]), exc_info=False)
                continue

            logger.info(
                "Running {} for {}".format(parser.__class__.__name__, data["@id"])
            )
            try:
                parser.run()
            except Exception as e:
                logger.warning(
                    "Error when running {} for {}: {!s}".format(
                        parser.__class__.__name__, data["@id"], e
                    ),
                    exc_info=False,
                )

            # Aggregate elements from children parsers into this parser's elements
            self.elements.extend(parser.elements)

        return i + 1

    def run(self):
        assert WORKER_RUN_ID, "A WorkerRun ID is required"
        try:
            self.create_folder()
            if self.folder:
                self.create_folder_metadata()
            # IIIF 2&3: members
            count = self._run_from("members")
            if count < 1:
                # Fallback for IIIF 2 only: manifests and collections
                self._run_from("manifests")
                self._run_from("collections")
        finally:
            self.close()


PARSERS = {"sc:Manifest": ManifestParser, "sc:Collection": CollectionParser}
