# -*- coding: utf-8 -*-
import logging
import tempfile
from pathlib import Path

from PIL import Image, ImageOps

# Image formats that should not be converted
# see https://pillow.readthedocs.io/en/5.1.x/handbook/image-file-formats.html
VALID_FORMATS = ["JPEG", "JPEG2000", "PNG"]

logger = logging.getLogger(__name__)


def check_image(datafile):
    try:
        img = Image.open(datafile["local_path"])
    except IOError:
        raise ValueError("File {} is not a valid image".format(datafile["name"]))

    # Keep track of whether or not the image needs to be re-saved
    changed = False

    img_path = Path(tempfile.mkdtemp()) / datafile["id"]
    img_format = img.format

    # Default to JPEG
    if img_format not in VALID_FORMATS:
        img_path = f"{img_path}.jpg"
        img_format = "JPEG"
        changed = True

    if img.getexif().get(0x0112, 0) not in (0, 1):
        logger.info(f"Image ({img.size}) is rotated, it will be transposed.")
        # Apply the conversion specified by the EXIF Orientation tag if it is set and is not 1
        # (1 is TopLeft, which is equivalent to doing nothing)
        img = ImageOps.exif_transpose(img)
        changed = True
        assert img.getexif().get(0x0112, 0) in (
            0,
            1,
        ), "Image should not be rotated anymore"

    # Convert to RGB if needed
    if img.mode != "RGB":
        img = img.convert("RGB")
        changed = True

    if changed:
        img.save(img_path, format=img_format)
        return [Path(img_path)]
    return [datafile["local_path"]]
