# -*- coding: utf-8 -*-
import logging

from shapely.geometry import LinearRing
from teklia_toolbox.pagexml import PageXmlPage
from teklia_toolbox.requests import retried_request

from arkindex_tasks.base import WORKER_RUN_ID

logger = logging.getLogger(__name__)

MAX_POLYGON_POINTS = 512


class PageXmlParser(object):
    def __init__(self, path_or_xml, image):
        self.pagexml_page = PageXmlPage(path_or_xml)
        self.image = image

    def simplify_polygon(self, region):
        if region.points is None:
            logger.warning("No points in region {}".format(region.id))
            return

        # Ignore polygons that are completely outside the image
        if not any(
            0 < x <= self.image["width"] and 0 < y <= self.image["height"]
            for x, y in region.points
        ):
            logger.warning(
                "Ignoring region {} (this polygon is entirely outside the image's bounds)".format(
                    region.id
                )
            )
            return

        # Crop polygons that are partially outside the image
        cropped_polygon = [
            [
                min(self.image["width"], max(0, x)),
                min(self.image["height"], max(0, y)),
            ]
            for x, y in region.points
        ]

        if len(cropped_polygon) < 3:
            logger.warning(
                f"Ignoring region {region.id} (not enough points in polygon)"
            )
            return

        polygon = LinearRing(cropped_polygon)

        for tolerance in range(1, 20):
            simplified = polygon.simplify(tolerance)
            points = list(simplified.coords)
            if len(points) <= MAX_POLYGON_POINTS:
                return points

        logger.warning(
            f"Ignoring region {region.id} (no points simplification possible using normal tolerances)"
        )

    def check_transcription(self, region):
        if region.text is None or not region.text.strip():
            logger.warning("No text in region {}".format(region.id))
            return False

        return True

    def save(
        self, corpus_id, parent_id, image_id, page_type, paragraph_type, line_type
    ):
        if WORKER_RUN_ID is None:
            logger.error("Worker run ID not found")
            return None, []

        # The metadata can be missing if `exportTranscriptMetadata` is set to `false` in an export
        if (
            self.pagexml_page.metadata is None
            or self.pagexml_page.metadata.transkribus_metadata is None
        ):
            logger.error("Missing TranskribusMetadata XML tag in transcript")
            return None, []

        element = retried_request(
            "CreateElement",
            body={
                "corpus": corpus_id,
                "type": page_type["slug"],
                "parent": parent_id,
                "name": str(
                    self.pagexml_page.metadata.transkribus_metadata.page_number
                ),
                "image": image_id,
                "worker_run_id": WORKER_RUN_ID,
            },
        )
        retried_request(
            "CreateMetaData",
            id=element["id"],
            body={
                "type": "reference",
                "name": "transkribus_page_id",
                "value": str(self.pagexml_page.metadata.transkribus_metadata.page_id),
                "worker_run_id": WORKER_RUN_ID,
            },
        )
        retried_request(
            "CreateMetaData",
            id=element["id"],
            body={
                "type": "text",
                "name": "folio",
                "value": str(
                    self.pagexml_page.metadata.transkribus_metadata.page_number
                ),
                "worker_run_id": WORKER_RUN_ID,
            },
        )
        retried_request(
            "CreateMetaData",
            id=element["id"],
            body={
                "type": "text",
                "name": "transkribus_status",
                "value": str(self.pagexml_page.metadata.transkribus_metadata.status),
                "worker_run_id": WORKER_RUN_ID,
            },
        )

        if self.pagexml_page.page.text_regions is None or not len(
            self.pagexml_page.page.text_regions
        ):
            logger.warning("No transcriptions to save")
            return element, []

        region_count, line_count, region_ts_count, line_ts_count = 0, 0, 0, 0

        # Create paragraphs from text regions in a single call
        region_payloads = [
            (
                region,
                {"type": paragraph_type["slug"], "name": str(i), "polygon": polygon},
            )
            for i, region in enumerate(self.pagexml_page.page.text_regions)
            if (polygon := self.simplify_polygon(region))
        ]

        if not region_payloads:
            logger.warning("No valid regions to save")
            return element, []

        logger.debug("Creating region elements")
        regions_response = retried_request(
            "CreateElements",
            id=element["id"],
            body={
                "worker_run_id": WORKER_RUN_ID,
                "elements": [payload for _, payload in region_payloads],
            },
        )
        region_count += len(regions_response)

        transcription_payloads = []

        # Add transcription payloads for regions with non-empty text
        for (region, _), region_element in zip(region_payloads, regions_response):
            if self.check_transcription(region):
                # Entities can be on several lines, so the reading order is important to merge entities correctly.
                # It is sent in the transcriptions.json file for the entity build task,
                # so we store it along the payload and we'll link it with the transcription ID after creating.
                tag_reading_order = next(
                    filter(lambda tag: tag.name == "readingOrder", region.tags)
                )
                index_region = int(tag_reading_order.index)

                transcription_payloads.append(
                    (
                        {
                            "element_id": region_element["id"],
                            "text": region.text,
                            "confidence": 1
                            if region.confidence is None
                            else region.confidence,
                        },
                        index_region,
                    )
                )
                region_ts_count += 1

            # Create line elements in a single call
            line_payloads = [
                (line, {"polygon": polygon, "type": line_type["slug"], "name": str(i)})
                for i, line in enumerate(region.lines)
                if (polygon := self.simplify_polygon(line))
            ]
            if line_payloads:
                logger.debug(f"Creating line elements in region {region.id}")
                lines_response = retried_request(
                    "CreateElements",
                    id=region_element["id"],
                    body={
                        "worker_run_id": WORKER_RUN_ID,
                        "elements": [payload for _, payload in line_payloads],
                    },
                )
                line_count += len(lines_response)

                for index, (line, _) in enumerate(line_payloads):
                    transkribus_id = line.get_text("@id")
                    if transkribus_id is not None:
                        retried_request(
                            "CreateMetaData",
                            id=lines_response[index]["id"],
                            body={
                                "type": "reference",
                                "name": "transkribus_id",
                                "value": transkribus_id,
                                "worker_run_id": WORKER_RUN_ID,
                            },
                        )

                # Add transcription payloads for lines with non-empty text
                for (line, _), line_element in zip(line_payloads, lines_response):
                    if self.check_transcription(line):
                        transcription_payloads.append(
                            (
                                {
                                    "element_id": line_element["id"],
                                    "text": line.text,
                                    "confidence": 1
                                    if line.confidence is None
                                    else line.confidence,
                                },
                                # No reading order here, as lines are ignored for entity building
                                None,
                            )
                        )
                        line_ts_count += 1

        # Create all transcriptions for the page's regions and lines at once
        transcriptions = []
        if transcription_payloads:
            logger.debug("Creating transcriptions")
            response = retried_request(
                "CreateTranscriptions",
                body={
                    "worker_run_id": WORKER_RUN_ID,
                    "transcriptions": [
                        payload for (payload, _) in transcription_payloads
                    ],
                },
            )
            # Build the transcriptions.json payload
            transcriptions = [
                {"index_region": index_region, "transcription_id": transcription["id"]}
                for (_, index_region), transcription in zip(
                    transcription_payloads, response["transcriptions"]
                )
                if index_region is not None
            ]

        logger.info(
            f'Created {region_count} {paragraph_type["slug"]} elements with {region_ts_count} transcriptions '
            f'and {line_count} {line_type["slug"]} elements with {line_ts_count} transcriptions.'
        )
        return element, transcriptions
