# -*- coding: utf-8 -*-
import distutils.spawn
import logging
import tempfile
from glob import glob
from pathlib import Path

from apistar.exceptions import ErrorResponse
from pdf2image import convert_from_path
from pdfminer.high_level import extract_pages
from pdfminer.layout import LTTextContainer
from pdfminer.pdfdocument import PDFDocument
from pdfminer.pdfpage import PDFPage
from pdfminer.pdfparser import PDFParser
from teklia_toolbox.time import Timer

from arkindex_tasks import default_client
from arkindex_tasks.base import WORKER_RUN_ID
from arkindex_tasks.import_files.pdf_language import get_text_orientation

logger = logging.getLogger(__name__)

# Silent debugging infos from pdfminer
pdf_logger = logging.getLogger("pdfminer")
pdf_logger.setLevel(logging.WARNING)


def extract_pdf_images(datafile, base_path=None):
    if datafile["content_type"] != "application/pdf":
        raise ValueError("File {} is not a PDF".format(datafile["name"]))

    logger.info("Converting PDF file {} using Poppler".format(datafile["name"]))
    assert distutils.spawn.find_executable("pdftoppm"), "Missing pdftoppm in PATH"

    if base_path is None:
        base_path = Path(tempfile.mkdtemp())
    logger.debug("Storing JPG files in {!s}".format(base_path))

    with Timer() as t:
        convert_from_path(
            datafile["local_path"],
            output_folder=base_path,
            output_file="pdf-",  # pdf- prefix on images
            dpi=300,
            fmt="jpg",
        )

    logger.info("Conversion took {!s}".format(t.delta))

    return sorted(glob(str(base_path / "pdf-*.jpg")))


def build_transcription(pdf_element, pdf_page, ark_page):
    if not pdf_element.get_text().strip():
        return

    image_width, image_height = (
        ark_page["zone"]["image"]["width"],
        ark_page["zone"]["image"]["height"],
    )
    ratio_w = image_width / pdf_page.width
    ratio_h = image_height / pdf_page.height

    # Invert the Y coordinates: they are given from the bottom of the page
    y0 = pdf_page.height - pdf_element.y0
    y1 = pdf_page.height - pdf_element.y1

    polygon = [
        [round(pdf_element.x0 * ratio_w), round(y0 * ratio_h)],
        [round(pdf_element.x1 * ratio_w), round(y0 * ratio_h)],
        [round(pdf_element.x1 * ratio_w), round(y1 * ratio_h)],
        [round(pdf_element.x0 * ratio_w), round(y1 * ratio_h)],
    ]
    # Clamp all the coordinates so they don't leave the image
    polygon = [
        [
            min(image_width, max(0, x)),
            min(image_height, max(0, y)),
        ]
        for x, y in polygon
    ]

    # Check that we have four different points in the polygon to avoid API errors
    if len(set(map(tuple, polygon))) < 4:
        return

    return {
        "text": pdf_element.get_text(),
        "polygon": polygon,
        "confidence": 1.0,
        "orientation": get_text_orientation(pdf_element),
    }


def extract_pdf_text(path, ark_pages, existing_pages=None):
    # Load all pages and children
    pdf_pages = list(extract_pages(path))
    assert len(pdf_pages) == len(
        ark_pages
    ), f"Invalid nb of pages: pdf has {len(pdf_pages)}, ark has {len(ark_pages)}"

    # Do not upload transcriptions for pages that already existed on Arkindex (retried imports)
    if not existing_pages:
        existing_pages = []
    out = {}
    for ark_page, pdf_page in zip(ark_pages, pdf_pages):
        if ark_page["id"] in existing_pages:
            continue
        logger.debug(
            f"PDF text extraction on arkindex element {ark_page['id']} and pdf page {pdf_page}"
        )

        texts = []
        for element in pdf_page:

            if isinstance(element, LTTextContainer):
                # Publish all lines directly
                texts += [
                    build_transcription(elt, pdf_page, ark_page)
                    for elt in element
                    if elt.get_text().strip()
                ]

            else:
                logger.debug(f"Skip pdf element {element}")

        transcriptions = list(filter(None, texts))
        if transcriptions:
            out[ark_page["id"]] = transcriptions

    return out


def count_pdf_pages(path) -> int:
    with open(path, "rb") as f:
        document = PDFDocument(PDFParser(f))
        # create_pages returns a generator, we use a sum to get its length while discarding its output
        return sum(1 for _ in PDFPage.create_pages(document))


def save_pdf_transcriptions(parent_id, transcriptions) -> None:
    assert WORKER_RUN_ID, "Missing WorkerRun ID"
    try:
        default_client.request(
            "CreateElementTranscriptions",
            id=parent_id,
            body={
                "element_type": "text_line",
                "worker_run_id": WORKER_RUN_ID,
                "transcriptions": transcriptions,
            },
        )
    except ErrorResponse as e:
        logger.warning(f"Transcriptions publication failed: {e.content}")
        raise


def upload_pdf_text(pdf_path, ark_pages, existing_pages=None) -> None:
    """
    Upload transcriptions from the text found in a PDF file to existing Arkindex elements.

    :param pdf_path: Path to the PDF file.
    :type pdf_path: str or pathlib.Path
    :param ark_pages list: List of existing Arkindex elements matching each page of the PDF,
       as they would be returned by the `ListElements` or `RetrieveElement` API endpoints.
    :param existing_pages: List of Arkindex elements that should be skipped,
       as they already have transcriptions.
    :type existing_pages: list or None
    """
    for page_id, transcriptions in extract_pdf_text(
        pdf_path, ark_pages, existing_pages=existing_pages
    ).items():
        save_pdf_transcriptions(page_id, transcriptions)
