#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import argparse
import hashlib
import logging
import os
import shutil
import tempfile
import uuid
from itertools import groupby
from pathlib import Path
from zipfile import ZipFile

import requests
from apistar.exceptions import ErrorResponse
from lxml import etree
from PIL import Image, ImageOps
from teklia_toolbox.requests import retried_request

from arkindex_tasks import default_client
from arkindex_tasks.base import WORKER_RUN_ID, dump_json
from arkindex_tasks.import_files.pagexml import PageXmlParser

logger = logging.getLogger(__name__)


def hash_file(path) -> str:
    md5 = hashlib.md5()
    with open(path, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            md5.update(chunk)
    return md5.hexdigest()


class TranskribusElement(object):
    def __init__(
        self,
        archive,
        image_filename,
        xml_filename,
        corpus_id,
        folder_id,
        element_type,
        paragraph_type,
        line_type,
    ):
        self.archive = archive
        self.image_filename = image_filename
        self.xml_filename = xml_filename
        self.corpus_id = corpus_id
        self.folder_id = folder_id
        self.element_type = element_type
        self.paragraph_type = paragraph_type
        self.line_type = line_type

    def save_image(self):
        logger.info(f"Extracting image {self.image_filename}")

        # Extract the image first before editing it.
        # Using Image.open(ZipFile.open()) is 15 times slower than extracting first!
        # Do not save in get_working_dir() here as it would become an artifact
        _, img_path = tempfile.mkstemp()
        with self.archive.open(self.image_filename) as source, open(
            img_path, "wb"
        ) as destination:
            shutil.copyfileobj(source, destination)

        with Image.open(img_path) as img:
            # Convert to RGB if needed
            if img.mode != "RGB":
                img = img.convert("RGB")

            # Apply the rotation/transposition from the EXIF orientation tag if it is used
            img = ImageOps.exif_transpose(img)
            img.save(img_path, format="JPEG")

        return img_path

    def upload_image(self):
        image_path = self.save_image()
        logger.info(f"Creating image {self.image_filename}")
        try:
            self.image = retried_request(
                "CreateImage", body={"hash": hash_file(image_path)}
            )
        except ErrorResponse as e:
            if e.status_code == 400 and "id" in e.content:
                self.image = retried_request("RetrieveImage", id=e.content["id"])
            else:
                raise
        else:
            logger.info(f"Uploading image {self.image_filename} as {self.image['id']}")
            try:
                with open(image_path, "rb") as f:
                    requests.put(self.image["s3_put_url"], data=f)
            finally:
                os.unlink(image_path)

        logger.info(f"Validating image {self.image['id']}")
        # Store the returned image as it might include updated dimensions
        self.image = retried_request(
            "PartialUpdateImage", id=self.image["id"], body={"status": "checked"}
        )

    def upload_transcriptions(self):
        element = None
        logger.info(f"Importing transcript {self.xml_filename}")
        try:
            # Read and parse the xml file
            xml_text = self.archive.read(self.xml_filename)
            parser = PageXmlParser(etree.fromstring(xml_text), self.image)
            element, _ = parser.save(
                corpus_id=self.corpus_id,
                parent_id=self.folder_id,
                image_id=self.image["id"],
                page_type=self.element_type,
                paragraph_type=self.paragraph_type,
                line_type=self.line_type,
            )

            if element is not None:
                retried_request(
                    "CreateMetaData",
                    id=element["id"],
                    body={
                        "type": "text",
                        "name": "original_image_name",
                        # Use the imageFilename XML attribute if available, fallback on the filename from the archive
                        "value": parser.pagexml_page.page.image_name
                        or os.path.basename(self.image_filename),
                        "worker_run_id": WORKER_RUN_ID,
                    },
                )
        except AssertionError as e:
            logger.error(f"Failed importing transcript {self.xml_filename}: {e}")
        except ErrorResponse as e:
            logger.error(
                f"Failed importing transcript {self.xml_filename}: {e.status_code} - {e.content}"
            )
        except requests.exceptions.RequestException as e:
            logger.error(f"Failed importing transcript {self.xml_filename}: {e}")
        return element

    def run(self):
        try:
            self.upload_image()
            return self.upload_transcriptions()
        except ErrorResponse as e:
            logger.error(
                f"Failed importing {self.image_filename} and {self.xml_filename}: {e.status_code} - {e.content}"
            )
        except Exception as e:
            logger.error(
                f"Failed importing {self.image_filename} and {self.xml_filename}: {e}"
            )
        return []


class TranskribusFolder(object):
    def __init__(
        self,
        archive,
        filenames,
        title,
        corpus_id,
        folder_type,
        element_type,
        paragraph_type,
        line_type,
        parent_id=None,
    ):
        self.archive = archive
        self.filenames = filenames
        self.title = title
        self.corpus_id = corpus_id
        self.parent_id = parent_id
        self.folder_type = folder_type
        self.element_type = element_type
        self.paragraph_type = paragraph_type
        self.line_type = line_type
        self.folder = None

    def get_or_create_folder(self):
        if self.parent_id is None:
            search = default_client.paginate(
                "ListElements",
                corpus=self.corpus_id,
                name=self.title,
                type=self.folder_type["slug"],
            )
        else:
            search = default_client.paginate(
                "ListElementChildren",
                id=self.parent_id,
                name=self.title,
                type=self.folder_type["slug"],
                recursive=True,
            )

        for folder in search:
            # The name filter on ListElements will match if the name of any element contains self.title (case insensitive)
            # so we need to check again for strict equality.
            if folder["name"] == self.title:
                self.folder = folder
                return False

        logger.info(f"Creating {self.folder_type['display_name']} {self.title}")
        self.folder = retried_request(
            "CreateElement",
            body={
                "corpus": self.corpus_id,
                "parent": self.parent_id,
                "type": self.folder_type["slug"],
                "name": self.title,
                "worker_run_id": WORKER_RUN_ID,
            },
        )
        return True

    def run(self):
        elements = []
        try:
            created = self.get_or_create_folder()
            elements.append(self.folder)
            if not created:
                logger.info(
                    "Skipping existing {} {}".format(
                        self.folder_type["display_name"], self.folder["name"]
                    )
                )
                return elements

            # Group files by name without extension: each group will hold the PAGE XML file and the image
            for name, filenames in groupby(
                sorted(self.filenames, key=lambda path: path.stem),
                key=lambda path: path.stem,
            ):
                # Ignore the METS file and the Transkribus metadata file
                if name in ("mets", "metadata"):
                    continue

                filenames = list(filenames)
                if len(filenames) != 2:
                    logger.warning(f"Expected 2 files for page {name}; skipping")
                    continue

                if filenames[0].suffix == ".xml":
                    xml_filename, image_filename = filenames
                else:
                    image_filename, xml_filename = filenames

                elements.append(
                    TranskribusElement(
                        archive=self.archive,
                        image_filename=str(image_filename),
                        xml_filename=str(xml_filename),
                        corpus_id=self.corpus_id,
                        folder_id=self.folder["id"],
                        element_type=self.element_type,
                        paragraph_type=self.paragraph_type,
                        line_type=self.line_type,
                    ).run()
                )
        except ErrorResponse as e:
            logger.error(
                "Failed importing {} {}: {} - {}".format(
                    self.folder_type["display_name"],
                    self.title,
                    e.status_code,
                    e.content,
                )
            )
        except Exception as e:
            logger.error(
                "Failed importing {} {}: {}".format(
                    self.folder_type["display_name"], self.title, e
                )
            )
        return elements


class TranskribusImporter(object):
    def __init__(
        self,
        archive_path,
        *,
        corpus,
        folder_type,
        element_type,
        parent_id=None,
        paragraph_type="paragraph",
        line_type="text_line",
    ):

        assert archive_path.is_file(), f"Archive at {archive_path} not found"
        self.archive_path = archive_path

        assert WORKER_RUN_ID, "A WorkerRun ID is required"

        # Check corpus
        try:
            self.corpus = retried_request("RetrieveCorpus", id=corpus)
        except ErrorResponse as e:
            if e.status_code == 404:
                logger.error("Corpus {} not found".format(corpus))
            raise

        # Check parent element
        if parent_id:
            try:
                self.parent = retried_request("RetrieveElement", id=parent_id)
            except ErrorResponse as e:
                if e.status_code == 404:
                    logger.error(f"Parent element {parent_id} not found")
                raise

            if self.parent["corpus"]["id"] != self.corpus["id"]:
                raise ValueError(
                    f"Parent element {parent_id} is not in corpus {self.corpus['id']}"
                )
        else:
            self.parent = None

        # Check folder type
        try:
            self.folder_type = next(
                t for t in self.corpus["types"] if t["slug"] == folder_type
            )
        except StopIteration:
            raise ValueError(
                "Type {} not found in corpus {}".format(
                    folder_type, self.corpus["name"]
                )
            )

        if not self.folder_type["folder"]:
            raise ValueError("Type {} is not a folder type".format(folder_type))

        # Check element type
        try:
            self.element_type = next(
                t for t in self.corpus["types"] if t["slug"] == element_type
            )
        except StopIteration:
            raise ValueError(
                "Type {} not found in corpus {}".format(
                    element_type, self.corpus["name"]
                )
            )

        if self.element_type["folder"]:
            raise ValueError("Type {} should not be a folder type".format(element_type))

        # Check paragraph type
        try:
            self.paragraph_type = next(
                t for t in self.corpus["types"] if t["slug"] == paragraph_type
            )
        except StopIteration:
            raise ValueError(
                "Type {} not found in corpus {}".format(
                    paragraph_type, self.corpus["name"]
                )
            )

        if self.paragraph_type["folder"]:
            raise ValueError(
                "Type {} should not be a folder type".format(paragraph_type)
            )

        # Check line type
        try:
            self.line_type = next(
                t for t in self.corpus["types"] if t["slug"] == line_type
            )
        except StopIteration:
            raise ValueError(
                "Type {} not found in corpus {}".format(line_type, self.corpus["name"])
            )

        if self.line_type["folder"]:
            raise ValueError("Type {} should not be a folder type".format(line_type))

    def run(self):
        elements = []
        # Archive structure: <document ID>/<document title>/[<optional subdir>/]<page number>_<page ID>.[xml,jpg,tif…]
        # Also includes a log.txt file at the root and a mets.xml in each document folder.
        with ZipFile(self.archive_path) as archive:

            paths = sorted(map(Path, archive.namelist()))

            def grouper(path):
                """
                Group all files by `docID/title`.
                Path('a/b/c/d').parents returns an iterator for the paths a/b/c, a/b, a, and `.`.
                With an absolute path, we would get `/` as the last parent instead of `.`.
                The folders that we are looking for are a/b, so we want the third-to-last item.
                We cannot access a parent by index before Python 3.10, so until then, we also use a list().
                """
                parents = list(path.parents)
                if len(parents) < 3:
                    # This is not a document directory.
                    return None
                return parents[-3]

            for parent, filenames in groupby(paths, key=grouper):
                # Ignore any file found outside of a document, such as `log.txt`.
                if not parent:
                    continue

                try:
                    elements += TranskribusFolder(
                        archive,
                        filenames,
                        # `parent` here is <doc ID>/<doc title>, so we extract the title
                        title=parent.name,
                        corpus_id=self.corpus["id"],
                        parent_id=self.parent["id"] if self.parent else None,
                        folder_type=self.folder_type,
                        element_type=self.element_type,
                        paragraph_type=self.paragraph_type,
                        line_type=self.line_type,
                    ).run()
                except Exception as e:
                    logger.error(f"Failed importing folder {parent}: {e}")

        return elements


def main():
    parser = argparse.ArgumentParser(
        description="Import a Transkribus ZIP export into Arkindex"
    )
    parser.add_argument(
        "archive_path",
        help="Path to a ZIP file holding a Transkribus export",
        type=Path,
    )
    parser.add_argument(
        "--corpus",
        help="UUID of an existing corpus to import into",
        type=uuid.UUID,
        required=True,
    )
    parser.add_argument(
        "--parent-id",
        help="UUID of an existing parent element in the corpus to import into",
        type=uuid.UUID,
    )
    parser.add_argument(
        "--folder-type",
        help="Slug of an element type to use for documents",
        default="volume",
    )
    parser.add_argument(
        "--element-type",
        help="Slug of an element type to use for pages",
        default="page",
    )
    parser.add_argument(
        "--paragraph-type",
        help="Slug of an element type to use for paragraph element",
        default="paragraph",
    )
    parser.add_argument(
        "--line-type",
        help="Slug of an element type to use for line element",
        default="text_line",
    )
    parser.add_argument(
        "--sleep",
        help="Throttle API requests by waiting for a given number of seconds",
        type=float,
        default=0,
    )
    args = vars(parser.parse_args())

    default_client.sleep_duration = args.pop("sleep")

    elements = TranskribusImporter(**args).run()
    dump_json(elements)


if __name__ == "__main__":
    main()
