# -*- coding: utf-8 -*-
import json
from io import BytesIO
from pathlib import Path
from unittest import TestCase
from unittest.mock import call, patch

import requests_mock

from arkindex_tasks.import_files.iiif import CollectionParser

SAMPLES = Path(__file__).absolute().parent.parent / "samples"


@requests_mock.Mocker()
class TestCollectionParser(TestCase):
    def test_create_folder_kwargs(self, req_mock):
        req_mock.post(
            "/api/v1/elements/create/",
            json={"id": "folderid", "name": "foldername", "type": "thing"},
        )
        parser = CollectionParser(
            BytesIO(b"{}"),
            first_level=False,
            corpus_id="corpusid",
            folder_type="thing",
            folder_name="foldername",
        )
        parser.create_folder()
        self.assertIn("parent_id", parser.parser_kwargs)
        self.assertEqual(parser.parser_kwargs["parent_id"], "folderid")
        self.assertEqual(req_mock.call_count, 1)

    def test_create_folder_first_level(self, req_mock):
        parser = CollectionParser(
            BytesIO(), first_level=True, corpus_id="corpusid", folder_type="thing"
        )
        parser.create_folder()
        self.assertEqual(req_mock.call_count, 0)
        self.assertIsNone(parser.parser_kwargs.get("parent_id"))

    @patch("arkindex_tasks.import_files.iiif.IIIFParser.load")
    def test_run_from(self, req_mock, load_mock):
        req_mock.get(
            "http://iiif/subcollection/", json={"@id": "http://iiif/subcollection/"}
        )
        req_mock.get("http://iiif/manifest/", json={"@id": "http://iiif/manifest/"})
        with (SAMPLES / "collection.json").open("rb") as f:
            parser = CollectionParser(f)
            self.assertEqual(parser._run_from("members"), 2)

        self.assertEqual(req_mock.call_count, 2)
        self.assertListEqual(
            [(req.method, req.url) for req in req_mock.request_history],
            [("GET", "http://iiif/subcollection/"), ("GET", "http://iiif/manifest/")],
        )

    @patch("arkindex_tasks.import_files.iiif.logger")
    @patch("arkindex_tasks.import_files.iiif.IIIFParser.load")
    def test_run_from_exceptions(self, req_mock, load_mock, logger_mock):
        req_mock.get("http://iiif/subcollection/", status_code=404)
        req_mock.get("http://iiif/manifest/", json={"@id": "http://iiif/manifest/"})
        load_mock().run.side_effect = ZeroDivisionError("Oh no")
        with (SAMPLES / "collection.json").open("rb") as f:
            parser = CollectionParser(f)
            self.assertEqual(parser._run_from("members"), 2)

        self.assertEqual(req_mock.call_count, 2)
        self.assertListEqual(
            [(req.method, req.url) for req in req_mock.request_history],
            [("GET", "http://iiif/subcollection/"), ("GET", "http://iiif/manifest/")],
        )

        self.assertEqual(logger_mock.warning.call_count, 2)
        self.assertListEqual(
            logger_mock.warning.call_args_list,
            [
                call(
                    "Failed fetching manifest/collection at http://iiif/subcollection/: "
                    "404 Client Error: None for url: http://iiif/subcollection/",
                    exc_info=False,
                ),
                call(
                    "Error when running MagicMock for http://iiif/manifest/: Oh no",
                    exc_info=False,
                ),
            ],
        )

    def test_run_from_not_found(self, req_mock):
        parser = CollectionParser(BytesIO(b"{}"))
        self.assertEqual(parser._run_from("something"), 0)

    @patch(
        "arkindex_tasks.import_files.iiif.WORKER_RUN_ID",
        "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
    )
    def test_run(self, req_mock):
        req_mock.get(
            "http://iiif/subcollection/", json={"@id": "http://iiif/subcollection/"}
        )
        req_mock.get("http://iiif/manifest/", json={"@id": "http://iiif/manifest/"})
        with (SAMPLES / "collection.json").open("rb") as f:
            parser = CollectionParser(f)
            parser.run()

        self.assertEqual(req_mock.call_count, 2)
        self.assertListEqual(
            [(req.method, req.url) for req in req_mock.request_history],
            [("GET", "http://iiif/subcollection/"), ("GET", "http://iiif/manifest/")],
        )

    @patch(
        "arkindex_tasks.import_files.iiif.WORKER_RUN_ID",
        "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
    )
    def test_run_fallback(self, req_mock):
        req_mock.get(
            "http://iiif/subcollection/fallback/",
            json={"@id": "http://iiif/subcollection/fallback/"},
        )
        req_mock.get(
            "http://iiif/manifest/fallback/",
            json={"@id": "http://iiif/manifest/fallback/"},
        )
        with (SAMPLES / "collection.json").open("rb") as f:
            data = json.load(f)

        # Remove `members` to force a fallback to `manifests` and `collections`
        del data["members"]
        parser = CollectionParser(BytesIO(json.dumps(data).encode("utf-8")))
        parser.run()

        self.assertEqual(req_mock.call_count, 2)
        self.assertListEqual(
            [(req.method, req.url) for req in req_mock.request_history],
            [
                ("GET", "http://iiif/manifest/fallback/"),
                ("GET", "http://iiif/subcollection/fallback/"),
            ],
        )
