# -*- coding: utf-8 -*-
from unittest import TestCase
from unittest.mock import call, patch

from arkindex_tasks.import_files.iiif import to_primitive, trim_warn


class TestIIIFHelpers(TestCase):
    def test_primitive_to_primitive(self):
        self.assertEqual(to_primitive("lol"), "lol")
        self.assertEqual(to_primitive(42), 42)
        self.assertTrue(to_primitive(True))
        self.assertIsNone(to_primitive(None))

    def test_dict_to_primitive(self):
        self.assertEqual(to_primitive({"a": "b", "@value": "something"}), "something")

    def test_list_to_primitive(self):
        self.assertEqual(
            to_primitive(
                [
                    {"a": "b", "@value": "something"},
                    {"c": "d", "@value": "somethingelse"},
                ]
            ),
            "something",
        )
        self.assertEqual(
            to_primitive(
                [
                    {"a": "b", "@value": "something"},
                    {"c": "d", "@language": "en-US", "@value": "somethingelse"},
                ]
            ),
            "somethingelse",
        )

    def test_invalid_objects_to_primitive(self):
        with self.assertRaises(ValueError):
            to_primitive({"a": "b"})

        with self.assertRaises(ValueError):
            to_primitive([{"a": "b"}, {"c": "d", "@value": "something"}])

        with self.assertRaises(ValueError):
            to_primitive([])

        with self.assertRaises(ValueError):
            to_primitive([{"@value": "something"}, 42, True])

    @patch("arkindex_tasks.import_files.iiif.logger")
    def test_trim_warn(self, logger_mock):
        self.assertEqual(trim_warn("A" * 42, name="Something", max_length=42), "A" * 42)
        self.assertEqual(logger_mock.warning.call_count, 0)

        self.assertEqual(
            trim_warn("A" * 1337, name="Something", max_length=42), "A" * 42
        )
        self.assertEqual(logger_mock.warning.call_count, 1)
        self.assertEqual(
            logger_mock.warning.call_args,
            call("Something is longer than the 42 character limit and will be trimmed"),
        )

    def test_many(self):
        self.assertListEqual(to_primitive("a", many=True), ["a"])
        self.assertListEqual(to_primitive(["a", "b"], many=True), ["a", "b"])
        self.assertListEqual(
            to_primitive([{"@value": "a"}, {"@value": "b"}], many=True), ["a", "b"]
        )
        self.assertListEqual(
            to_primitive(
                [{"@value": "a"}, {"@value": "b", "@language": "cn"}], many=True
            ),
            ["a", "b"],
        )
        self.assertListEqual(
            to_primitive(
                [
                    {"@value": "a", "@language": "en"},
                    {"@value": "b", "@language": "cn"},
                ],
                many=True,
            ),
            ["a"],
        )
