# -*- coding: utf-8 -*-
import json
from io import BytesIO, StringIO
from unittest import TestCase
from unittest.mock import patch

import requests_mock

from arkindex_tasks.import_files.iiif import IIIFParser, ManifestParser


class TestIIIFParser(TestCase):
    def test_stream_binary(self):
        """
        Check IIIFParser rejects streams opened in text mode
        """
        with self.assertRaises(
            AssertionError, msg="Stream should be a file-like object in binary mode"
        ):
            IIIFParser(StringIO("something"))

    def test_close(self):
        stream = BytesIO()
        IIIFParser(stream).close()
        self.assertTrue(stream.closed)

    def test_v3(self):
        stream = BytesIO(
            json.dumps(
                {
                    "@context": "http://iiif.io/api/presentation/3/context.json",
                    "id": "something",
                    "type": "sc:Manifest",
                }
            ).encode("utf-8")
        )
        with self.assertRaises(
            ValueError,
            msg="The import functionality does not support IIIF Presentation API version 3",
        ):
            IIIFParser(stream).check_version()

    def test_get_type(self):
        stream = BytesIO(
            json.dumps({"@id": "something", "@type": "sc:Manifest"}).encode("utf-8")
        )
        self.assertEqual(IIIFParser(stream).get_type(), ManifestParser)

    def test_get_type_invalid(self):
        with self.assertRaises(ValueError, msg="Missing @type property in JSON data"):
            IIIFParser(BytesIO(b"{}")).get_type()

    def test_get_type_unsupported(self):
        with self.assertRaises(ValueError, msg='Unsupported @type "sc:Fest"'):
            stream = BytesIO(
                json.dumps({"@id": "something", "@type": "sc:Fest"}).encode("utf-8")
            )
            IIIFParser(stream).get_type()

    def test_load(self):
        stream = BytesIO(
            json.dumps(
                {
                    "@id": "something",
                    "@type": "sc:Manifest",
                }
            ).encode("utf-8")
        )
        self.assertIsInstance(IIIFParser.load(stream), ManifestParser)

    def test_run(self):
        with self.assertRaises(NotImplementedError):
            IIIFParser(BytesIO()).run()

    @requests_mock.Mocker()
    @patch(
        "arkindex_tasks.import_files.iiif.WORKER_RUN_ID",
        "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
    )
    def test_create_element(self, req_mock):
        req_mock.post(
            "/api/v1/elements/create/",
            json={"id": "elementid", "name": "elementname", "type": "element"},
        )

        parser = IIIFParser(stream=BytesIO(b""), corpus_id="corpus_id")
        element = parser.create_element({"type": "element", "name": "elementname"})

        self.assertDictEqual(
            element, {"id": "elementid", "name": "elementname", "type": "element"}
        )
        self.assertListEqual(parser.elements, [{"id": "elementid", "type": "element"}])
        self.assertEqual(req_mock.call_count, 1)
        request = req_mock.request_history[0]
        self.assertEqual(
            request.url, "https://arkindex.teklia.com/api/v1/elements/create/"
        )
        self.assertEqual(request.method, "POST")
        body = json.loads(request.body.decode("utf-8"))
        self.assertDictEqual(body, {"type": "element", "name": "elementname"})

    @requests_mock.Mocker()
    @patch(
        "arkindex_tasks.import_files.iiif.WORKER_RUN_ID",
        "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
    )
    def test_create_element_iiif_id(self, req_mock):
        req_mock.post(
            "/api/v1/elements/create/",
            json={"id": "elementid", "name": "elementname", "type": "element"},
        )
        req_mock.post("/api/v1/element/elementid/metadata/", status_code=201)

        parser = IIIFParser(stream=BytesIO(b""), corpus_id="corpus_id")
        element = parser.create_element(
            {"type": "element", "name": "elementname"}, iiif_id="something"
        )

        self.assertDictEqual(
            element, {"id": "elementid", "name": "elementname", "type": "element"}
        )
        self.assertEqual(req_mock.call_count, 2)
        request = req_mock.request_history[1]
        self.assertEqual(
            request.url,
            "https://arkindex.teklia.com/api/v1/element/elementid/metadata/",
        )
        self.assertEqual(request.method, "POST")
        body = json.loads(request.body.decode("utf-8"))
        self.assertDictEqual(
            body,
            {
                "name": "IIIF ID",
                "type": "reference",
                "value": "something",
                "worker_run_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            },
        )

    @requests_mock.Mocker()
    @patch(
        "arkindex_tasks.import_files.iiif.WORKER_RUN_ID",
        "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
    )
    def test_create_folder(self, req_mock):
        req_mock.post(
            "/api/v1/elements/create/",
            json={"id": "folderid", "name": "foldername", "type": "volume"},
        )
        req_mock.post("/api/v1/element/folderid/metadata/", status_code=201)

        parser = IIIFParser(
            stream=BytesIO(
                json.dumps(
                    {"@id": "something", "@type": "sc:Manifest", "label": "Manifesto"}
                ).encode("utf-8")
            ),
            corpus_id="corpus_id",
            folder_name="Le Manifest",
            folder_type="volume",
        )
        parser.create_folder()

        self.assertDictEqual(
            parser.folder, {"id": "folderid", "name": "foldername", "type": "volume"}
        )
        self.assertListEqual(parser.elements, [{"id": "folderid", "type": "volume"}])
        self.assertEqual(req_mock.call_count, 2)
        request = req_mock.request_history[0]
        self.assertEqual(
            request.url, "https://arkindex.teklia.com/api/v1/elements/create/"
        )
        self.assertEqual(request.method, "POST")
        body = json.loads(request.body.decode("utf-8"))
        self.assertDictEqual(
            body,
            {
                "corpus": "corpus_id",
                "type": "volume",
                "name": "Le Manifest",
                "worker_run_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            },
        )
        request = req_mock.request_history[1]
        self.assertEqual(
            request.url, "https://arkindex.teklia.com/api/v1/element/folderid/metadata/"
        )
        self.assertEqual(request.method, "POST")
        body = json.loads(request.body.decode("utf-8"))
        self.assertDictEqual(
            body,
            {
                "name": "IIIF ID",
                "type": "reference",
                "value": "something",
                "worker_run_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            },
        )

    @requests_mock.Mocker()
    @patch(
        "arkindex_tasks.import_files.iiif.WORKER_RUN_ID",
        "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
    )
    def test_create_folder_parent(self, req_mock):
        req_mock.post(
            "/api/v1/elements/create/",
            json={"id": "folderid", "name": "foldername", "type": "volume"},
        )
        req_mock.post("/api/v1/element/folderid/metadata/", status_code=201)

        parser = IIIFParser(
            stream=BytesIO(
                json.dumps(
                    {"@id": "something", "@type": "sc:Manifest", "label": "Manifesto"}
                ).encode("utf-8")
            ),
            corpus_id="corpus_id",
            parent_id="parent_id",
            folder_name="Le Manifest",
            folder_type="volume",
        )
        parser.create_folder()

        self.assertDictEqual(
            parser.folder, {"id": "folderid", "name": "foldername", "type": "volume"}
        )
        self.assertListEqual(parser.elements, [{"id": "folderid", "type": "volume"}])
        self.assertEqual(req_mock.call_count, 2)
        request = req_mock.request_history[0]
        self.assertEqual(
            request.url, "https://arkindex.teklia.com/api/v1/elements/create/"
        )
        self.assertEqual(request.method, "POST")
        body = json.loads(request.body.decode("utf-8"))
        self.assertDictEqual(
            body,
            {
                "corpus": "corpus_id",
                "parent": "parent_id",
                "type": "volume",
                "name": "Le Manifest",
                "worker_run_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            },
        )

    @requests_mock.Mocker()
    @patch(
        "arkindex_tasks.import_files.iiif.WORKER_RUN_ID",
        "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
    )
    def test_create_folder_label(self, req_mock):
        req_mock.post(
            "/api/v1/elements/create/",
            json={"id": "folderid", "name": "foldername", "type": "volume"},
        )
        req_mock.post("/api/v1/element/folderid/metadata/", status_code=201)

        parser = IIIFParser(
            stream=BytesIO(
                json.dumps(
                    {"@id": "something", "@type": "sc:Manifest", "label": "Manifesto"}
                ).encode("utf-8")
            ),
            corpus_id="corpus_id",
            folder_type="volume",
        )
        parser.create_folder()

        self.assertDictEqual(
            parser.folder, {"id": "folderid", "name": "foldername", "type": "volume"}
        )
        self.assertListEqual(parser.elements, [{"id": "folderid", "type": "volume"}])
        self.assertEqual(req_mock.call_count, 2)
        request = req_mock.request_history[0]
        self.assertEqual(
            request.url, "https://arkindex.teklia.com/api/v1/elements/create/"
        )
        self.assertEqual(request.method, "POST")
        body = json.loads(request.body.decode("utf-8"))
        self.assertDictEqual(
            body,
            {
                "corpus": "corpus_id",
                "type": "volume",
                "name": "Manifesto",
                "worker_run_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            },
        )

    @requests_mock.Mocker()
    @patch(
        "arkindex_tasks.import_files.iiif.WORKER_RUN_ID",
        "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
    )
    def test_create_folder_filename(self, req_mock):
        req_mock.post(
            "/api/v1/elements/create/",
            json={"id": "folderid", "name": "foldername", "type": "volume"},
        )
        req_mock.post("/api/v1/element/folderid/metadata/", status_code=201)
        stream = BytesIO(
            json.dumps({"@id": "something", "@type": "sc:Manifest"}).encode("utf-8")
        )
        stream.name = "/path/to/el-manifesto.json"

        parser = IIIFParser(stream=stream, corpus_id="corpus_id", folder_type="volume")
        parser.create_folder()

        self.assertDictEqual(
            parser.folder, {"id": "folderid", "name": "foldername", "type": "volume"}
        )
        self.assertListEqual(parser.elements, [{"id": "folderid", "type": "volume"}])
        self.assertEqual(req_mock.call_count, 2)
        request = req_mock.request_history[0]
        self.assertEqual(
            request.url, "https://arkindex.teklia.com/api/v1/elements/create/"
        )
        self.assertEqual(request.method, "POST")
        body = json.loads(request.body.decode("utf-8"))
        self.assertDictEqual(
            body,
            {
                "corpus": "corpus_id",
                "type": "volume",
                "name": "el-manifesto",
                "worker_run_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            },
        )

    def test_create_folder_no_name(self):
        parser = IIIFParser(
            stream=BytesIO(
                json.dumps({"@id": "something", "@type": "sc:Manifest"}).encode("utf-8")
            ),
            corpus_id="corpus_id",
            folder_type="volume",
        )
        with self.assertRaises(AssertionError, msg="Could not determine a folder name"):
            parser.create_folder()
