# -*- coding: utf-8 -*-
import json
from io import BytesIO
from pathlib import Path
from unittest import TestCase
from unittest.mock import patch

import requests_mock

from arkindex_tasks.import_files.iiif import ManifestParser

SAMPLES = Path(__file__).absolute().parent.parent / "samples"


@patch(
    "arkindex_tasks.import_files.utils.retried_iiif_get.retry.wait.wait_fixed",
    new=0,
)
class TestManifestParser(TestCase):
    @requests_mock.Mocker()
    @patch(
        "arkindex_tasks.import_files.iiif.WORKER_RUN_ID",
        "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
    )
    def test_create_folder_metadata(self, req_mock):
        req_mock.post("/api/v1/element/folderid/metadata/", status_code=201)
        parser = ManifestParser(
            stream=BytesIO(
                json.dumps(
                    {
                        "@id": "something",
                        "metadata": [
                            {},
                            {"label": "nope"},
                            {"label": "nope", "value": ""},
                            {"value": "nope"},
                            {"label": "meta", "value": "data"},
                        ],
                    }
                ).encode("utf-8")
            )
        )
        parser.folder = {"id": "folderid", "name": "foldername"}
        parser.create_folder_metadata()

        self.assertEqual(req_mock.call_count, 1)
        request = req_mock.request_history[0]
        self.assertEqual(request.method, "POST")
        self.assertEqual(
            request.url, "https://arkindex.teklia.com/api/v1/element/folderid/metadata/"
        )
        body = json.loads(request.body.decode("utf-8"))
        self.assertDictEqual(
            body,
            {
                "name": "meta",
                "type": "text",
                "value": "data",
                "worker_run_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            },
        )

    @requests_mock.Mocker()
    @patch(
        "arkindex_tasks.import_files.iiif.WORKER_RUN_ID",
        "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
    )
    def test_create_metadata(self, req_mock):
        req_mock.post("/api/v1/element/elementid/metadata/", status_code=201)
        parser = ManifestParser(BytesIO())
        parser.create_metadata(
            element={"id": "elementid", "name": "An element"},
            metadatas=[{"label": "some label", "value": "some value"}],
        )
        self.assertEqual(req_mock.call_count, 1)
        request = req_mock.request_history[0]
        self.assertEqual(request.method, "POST")
        self.assertEqual(
            request.url,
            "https://arkindex.teklia.com/api/v1/element/elementid/metadata/",
        )
        body = json.loads(request.body.decode("utf-8"))
        self.assertDictEqual(
            body,
            {
                "name": "some label",
                "type": "text",
                "value": "some value",
                "worker_run_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            },
        )

    @requests_mock.Mocker()
    @patch(
        "arkindex_tasks.import_files.iiif.WORKER_RUN_ID",
        "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
    )
    def test_create_metadata_url(self, req_mock):
        req_mock.post("/api/v1/element/elementid/metadata/", status_code=201)
        parser = ManifestParser(BytesIO())
        parser.create_metadata(
            element={"id": "elementid", "name": "An element"},
            metadatas=[
                {"label": "not a url", "value": "gopher://arkindex.teklia.com:70/1/"},
                {"label": "a url", "value": "https://arkindex.teklia.com/"},
            ],
        )
        self.assertEqual(req_mock.call_count, 2)
        text_request, url_request = req_mock.request_history
        self.assertEqual(text_request.method, "POST")
        self.assertEqual(url_request.method, "POST")
        self.assertEqual(
            url_request.url,
            "https://arkindex.teklia.com/api/v1/element/elementid/metadata/",
        )
        self.assertEqual(
            text_request.url,
            "https://arkindex.teklia.com/api/v1/element/elementid/metadata/",
        )
        text_body = json.loads(text_request.body.decode("utf-8"))
        url_body = json.loads(url_request.body.decode("utf-8"))
        self.assertDictEqual(
            text_body,
            {
                "name": "not a url",
                "type": "text",
                "value": "gopher://arkindex.teklia.com:70/1/",
                "worker_run_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            },
        )
        self.assertDictEqual(
            url_body,
            {
                "name": "a url",
                "type": "url",
                "value": "https://arkindex.teklia.com/",
                "worker_run_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            },
        )

    @requests_mock.Mocker()
    def test_get_image(self, req_mock):
        req_mock.get("http://imageurl/info.json", json={"@id": "http://imageurl/"})
        req_mock.post(
            "/api/v1/image/iiif/info/",
            json={"id": "imageid", "url": "http://imageurl", "status": "checked"},
        )
        parser = ManifestParser(BytesIO())
        self.assertDictEqual(
            parser.get_image("http://imageurl"),
            {"id": "imageid", "url": "http://imageurl", "status": "checked"},
        )
        self.assertEqual(req_mock.call_count, 2)
        request = req_mock.request_history[0]
        self.assertEqual(request.method, "GET")
        self.assertEqual(request.url, "http://imageurl/info.json")
        request = req_mock.request_history[1]
        self.assertEqual(request.method, "POST")
        self.assertEqual(
            request.url, "https://arkindex.teklia.com/api/v1/image/iiif/info/"
        )
        body = json.loads(request.body.decode("utf-8"))
        self.assertDictEqual(body, {"@id": "http://imageurl/"})

    @requests_mock.Mocker()
    def test_get_image_check_error(self, req_mock):
        req_mock.get("http://imageurl/info.json", json={"@id": "http://imageurl/"})
        req_mock.post(
            "/api/v1/image/iiif/info/",
            status_code=400,
            json={"status": ["Image check failed!!1!"], "id": "imageid"},
        )
        req_mock.get(
            "/api/v1/image/imageid/",
            json={"id": "imageid", "url": "http://imageurl", "status": "error"},
        )
        parser = ManifestParser(BytesIO())
        self.assertDictEqual(
            parser.get_image("http://imageurl"),
            {"id": "imageid", "url": "http://imageurl", "status": "error"},
        )
        self.assertEqual(req_mock.call_count, 3)
        info_request, post_request, get_request = req_mock.request_history
        self.assertEqual(info_request.method, "GET")
        self.assertEqual(info_request.url, "http://imageurl/info.json")
        self.assertEqual(post_request.method, "POST")
        self.assertEqual(
            post_request.url, "https://arkindex.teklia.com/api/v1/image/iiif/info/"
        )
        body = json.loads(post_request.body.decode("utf-8"))
        self.assertDictEqual(body, {"@id": "http://imageurl/"})
        self.assertEqual(get_request.method, "GET")
        self.assertEqual(
            get_request.url, "https://arkindex.teklia.com/api/v1/image/imageid/"
        )

    @requests_mock.Mocker()
    def test_get_image_epic_fail(self, req_mock):
        req_mock.get("http://imageurl/info.json", status_code=500)
        parser = ManifestParser(BytesIO())
        self.assertIsNone(parser.get_image("http://imageurl"))
        self.assertEqual(req_mock.call_count, 2)
        info_request, info_request_retry = req_mock.request_history
        self.assertEqual(info_request.method, "GET")
        self.assertEqual(info_request.url, "http://imageurl/info.json")
        self.assertEqual(info_request_retry.method, "GET")
        self.assertEqual(info_request_retry.url, "http://imageurl/info.json")

    @requests_mock.Mocker()
    @patch(
        "arkindex_tasks.import_files.iiif.WORKER_RUN_ID",
        "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
    )
    def test_parse_canvas(self, req_mock):
        req_mock.get("http://imageurl/info.json", json={"@id": "http://imageurl"})
        req_mock.post(
            "/api/v1/image/iiif/info/",
            json={"id": "imageid", "url": "http://imageurl", "status": "checked"},
        )
        req_mock.post(
            "/api/v1/elements/create/",
            json={"id": "elementid", "name": "Thing 1", "type": "thing"},
        )
        parser = ManifestParser(BytesIO(), corpus_id="corpusid", element_type="thing")
        parser.folder = {"id": "folderid", "name": "foldername"}

        parser.parse_canvas(
            {
                "label": "Thing 1",
                "images": [{"resource": {"service": {"@id": "http://imageurl"}}}],
            }
        )

        self.assertListEqual(parser.elements, [{"id": "elementid", "type": "thing"}])
        self.assertEqual(req_mock.call_count, 3)
        info_request, image_request, element_request = req_mock.request_history
        self.assertEqual(info_request.method, "GET")
        self.assertEqual(info_request.url, "http://imageurl/info.json")
        self.assertEqual(image_request.method, "POST")
        self.assertEqual(
            image_request.url, "https://arkindex.teklia.com/api/v1/image/iiif/info/"
        )
        body = json.loads(image_request.body.decode("utf-8"))
        self.assertDictEqual(body, {"@id": "http://imageurl"})
        self.assertEqual(element_request.method, "POST")
        self.assertEqual(
            element_request.url, "https://arkindex.teklia.com/api/v1/elements/create/"
        )
        body = json.loads(element_request.body.decode("utf-8"))
        self.assertDictEqual(
            body,
            {
                "corpus": "corpusid",
                "parent": "folderid",
                "type": "thing",
                "name": "Thing 1",
                "image": "imageid",
                "worker_run_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            },
        )

    def test_parse_canvas_no_name(self):
        parser = ManifestParser(BytesIO())
        with self.assertRaises(ValueError, msg="Canvas has no name"):
            parser.parse_canvas({})

    def test_parse_canvas_no_image(self):
        parser = ManifestParser(BytesIO())
        with self.assertRaises(ValueError, msg="Canvas My canvas has no image"):
            parser.parse_canvas({"label": "My canvas", "images": []})

    def test_parse_canvas_no_resource(self):
        parser = ManifestParser(BytesIO())
        with self.assertRaises(
            ValueError, msg="Canvas My canvas image has no resource"
        ):
            parser.parse_canvas({"label": "My canvas", "images": [{}]})

    def test_parse_canvas_no_service(self):
        parser = ManifestParser(BytesIO())
        with self.assertRaises(
            ValueError, msg="Image resource A has no IIIF image information service"
        ):
            parser.parse_canvas(
                {"label": "My canvas", "images": [{"resource": {"@id": "A"}}]}
            )

    def test_parse_canvas_no_service_id(self):
        parser = ManifestParser(BytesIO())
        with self.assertRaises(
            ValueError, msg="Image service on resource A has no @id"
        ):
            parser.parse_canvas(
                {
                    "label": "My canvas",
                    "images": [{"resource": {"@id": "A", "service": {}}}],
                }
            )

    @requests_mock.Mocker()
    @patch(
        "arkindex_tasks.import_files.iiif.WORKER_RUN_ID",
        "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
    )
    def test_parse_canvas_image_error(self, req_mock):
        req_mock.get("http://imageurl/info.json", status_code=500)
        req_mock.post(
            "/api/v1/elements/create/",
            json={"id": "elementid", "name": "Thing 1", "type": "thing"},
        )
        parser = ManifestParser(BytesIO(), corpus_id="corpusid", element_type="thing")
        parser.folder = {"id": "folderid", "name": "foldername"}

        parser.parse_canvas(
            {
                "label": "Thing 1",
                "images": [{"resource": {"service": {"@id": "http://imageurl"}}}],
            }
        )

        self.assertListEqual(parser.elements, [{"id": "elementid", "type": "thing"}])
        self.assertEqual(req_mock.call_count, 3)
        info_request, info_request_retry, element_request = req_mock.request_history
        self.assertEqual(info_request.method, "GET")
        self.assertEqual(info_request.url, "http://imageurl/info.json")
        self.assertEqual(info_request_retry.method, "GET")
        self.assertEqual(info_request_retry.url, "http://imageurl/info.json")
        self.assertEqual(element_request.method, "POST")
        self.assertEqual(
            element_request.url, "https://arkindex.teklia.com/api/v1/elements/create/"
        )
        body = json.loads(element_request.body.decode("utf-8"))
        self.assertDictEqual(
            body,
            {
                "corpus": "corpusid",
                "parent": "folderid",
                "type": "thing",
                "name": "Thing 1",
                "image": None,
                "worker_run_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            },
        )

    @requests_mock.Mocker()
    def test_parse_canvases(self, req_mock):
        req_mock.get(
            "http://iiif/image1/info.json", json={"@id": "http://iiif/image1/"}
        )
        req_mock.get(
            "http://iiif/image2/info.json", json={"@id": "http://iiif/image2/"}
        )
        req_mock.post(
            "/api/v1/image/iiif/info/",
            [
                {
                    "json": {
                        "id": "image1",
                        "url": "http://iiif/image1/",
                        "status": "checked",
                    }
                },
                {
                    "json": {
                        "id": "image2",
                        "url": "http://iiif/image2/",
                        "status": "checked",
                    }
                },
            ],
        )
        req_mock.post(
            "/api/v1/elements/create/",
            [
                {"json": {"id": "element1", "name": "Thing 1", "type": "thing"}},
                {"json": {"id": "element2", "name": "Thing 2", "type": "thing"}},
            ],
        )
        req_mock.post("/api/v1/element/element2/metadata/", status_code=201)
        parser = ManifestParser(
            stream=(SAMPLES / "manifest.json").open("rb"),
            corpus_id="corpusid",
            element_type="thing",
        )
        parser.folder = {"id": "folderid", "name": "foldername"}
        try:
            parser.parse_canvases()
        finally:
            parser.close()

        self.assertListEqual(
            parser.elements,
            [{"id": "element1", "type": "thing"}, {"id": "element2", "type": "thing"}],
        )
        self.assertEqual(req_mock.call_count, 7)
        self.assertListEqual(
            [(req.method, req.url) for req in req_mock.request_history],
            [
                ("GET", "http://iiif/image1/info.json"),
                ("POST", "https://arkindex.teklia.com/api/v1/image/iiif/info/"),
                ("POST", "https://arkindex.teklia.com/api/v1/elements/create/"),
                ("GET", "http://iiif/image2/info.json"),
                ("POST", "https://arkindex.teklia.com/api/v1/image/iiif/info/"),
                ("POST", "https://arkindex.teklia.com/api/v1/elements/create/"),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/element2/metadata/",
                ),
            ],
        )

    @requests_mock.Mocker()
    @patch(
        "arkindex_tasks.import_files.iiif.WORKER_RUN_ID",
        "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
    )
    def test_run(self, req_mock):
        req_mock.post("/api/v1/element/folderid/metadata/", status_code=201)
        req_mock.post("/api/v1/element/element1/metadata/", status_code=201)
        req_mock.post("/api/v1/element/element2/metadata/", status_code=201)
        req_mock.get(
            "http://iiif/image1/info.json", json={"@id": "http://iiif/image1/"}
        )
        req_mock.get(
            "http://iiif/image2/info.json", json={"@id": "http://iiif/image2/"}
        )
        req_mock.post(
            "/api/v1/image/iiif/info/",
            [
                {
                    "status_code": 400,
                    "json": {"status": ["Image check failed!!1!"], "id": "image1"},
                },
                {
                    "json": {
                        "id": "image2",
                        "url": "http://iiif/image2/",
                        "status": "checked",
                    }
                },
            ],
        )
        req_mock.get(
            "/api/v1/image/image1/",
            json={"id": "image1", "url": "http://iiif/image1/", "status": "checked"},
        )
        req_mock.post(
            "/api/v1/elements/create/",
            [
                {"json": {"id": "folderid", "name": "foldername", "type": "folder"}},
                {"json": {"id": "element1", "name": "Thing 1", "type": "thing"}},
                {"json": {"id": "element2", "name": "Thing 2", "type": "thing"}},
            ],
        )
        parser = ManifestParser(
            stream=(SAMPLES / "manifest.json").open("rb"),
            corpus_id="corpusid",
            parent_id="parentid",
            folder_type="box",
            element_type="thing",
        )
        parser.run()

        self.assertListEqual(
            parser.elements,
            [
                {"id": "folderid", "type": "folder"},
                {"id": "element1", "type": "thing"},
                {"id": "element2", "type": "thing"},
            ],
        )
        self.assertEqual(req_mock.call_count, 12)
        self.assertListEqual(
            [(req.method, req.url) for req in req_mock.request_history],
            [
                ("POST", "https://arkindex.teklia.com/api/v1/elements/create/"),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/folderid/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/folderid/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/folderid/metadata/",
                ),
                ("GET", "http://iiif/image1/info.json"),
                ("POST", "https://arkindex.teklia.com/api/v1/image/iiif/info/"),
                ("GET", "https://arkindex.teklia.com/api/v1/image/image1/"),
                ("POST", "https://arkindex.teklia.com/api/v1/elements/create/"),
                ("GET", "http://iiif/image2/info.json"),
                ("POST", "https://arkindex.teklia.com/api/v1/image/iiif/info/"),
                ("POST", "https://arkindex.teklia.com/api/v1/elements/create/"),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/element2/metadata/",
                ),
            ],
        )
