# -*- coding: utf-8 -*-
import json
import tarfile
from datetime import datetime
from io import BufferedReader, BytesIO
from pathlib import Path
from tempfile import NamedTemporaryFile, TemporaryFile
from unittest import TestCase
from unittest.mock import call, patch
from zipfile import ZipFile

import requests_mock
from apistar.exceptions import ErrorResponse
from requests.exceptions import ConnectionError
from zstandard import ZstdCompressor

from arkindex_tasks.base import get_working_dir
from arkindex_tasks.import_files.base import FileImport

SAMPLES = Path(__file__).absolute().parent / "samples"


@requests_mock.Mocker()
@patch(
    "arkindex_tasks.import_files.base.FileImport.create_element.retry.wait.wait_fixed",
    new=0,
)
@patch(
    "teklia_toolbox.requests.download_file.retry.wait.wait_fixed",
    new=0,
)
@patch(
    "arkindex_tasks.import_files.base.FileImport.upload_image.retry.wait.wait_fixed",
    new=0,
)
class TestFileImport(TestCase):
    @patch("arkindex_tasks.import_files.base.check_image")
    def test_get_images(self, mock, check_image_mock):
        check_image_mock.return_value = [Path("/something.jpg")]
        test_datafile = {
            "id": "file1",
            "name": "file1.jpg",
            "local_path": "some/path/file1.jpg",
            "content_type": "image/jpeg",
        }

        fi = FileImport("processid")
        fi.process = {
            "id": "processid",
            "mode": "files",
            "element_type": "page",
            "element": {"id": "folderid"},
        }
        fi.get_images(test_datafile)

        self.assertEqual(check_image_mock.call_count, 1)
        self.assertEqual(check_image_mock.call_args, call(test_datafile))

    @patch("arkindex_tasks.import_files.base.extract_pdf_images")
    def test_get_images_pdf(self, mock, extract_pdf_images_mock):
        extract_pdf_images_mock.return_value = [Path("/something.jpg")]
        test_datafile = {
            "id": "file1",
            "name": "file1.pdf",
            "local_path": "some/path/file1.pdf",
            "content_type": "application/pdf",
        }

        fi = FileImport("processid")
        fi.process = {
            "id": "processid",
            "mode": "files",
            "element_type": "page",
            "element": {"id": "folderid"},
        }
        fi.get_images(test_datafile)

        self.assertEqual(extract_pdf_images_mock.call_count, 1)
        self.assertEqual(extract_pdf_images_mock.call_args, call(test_datafile))

    @patch("arkindex_tasks.import_files.base.check_image")
    def test_get_images_error(self, mock, check_image_mock):
        check_image_mock.side_effect = FileNotFoundError

        fi = FileImport("processid")
        fi.process = {
            "id": "processid",
            "mode": "files",
            "element_type": "page",
            "element": {"id": "folderid"},
        }
        datafile = {"id": "file1", "name": "file.jpg", "content_type": "image/jpeg"}
        self.assertListEqual(fi.get_images(datafile), [])

        self.assertEqual(check_image_mock.call_count, 1)
        self.assertEqual(check_image_mock.call_args, call(datafile))

    def test_create_image(self, mock):
        mock.post(
            "/api/v1/image/", json={"id": "imageid", "s3_put_url": "http://s3/put"}
        )
        mock.put("http://s3/put", status_code=201)
        mock.patch("/api/v1/image/imageid/", status_code=200)

        fi = FileImport("processid")
        img = fi.create_image(SAMPLES / "200x200.jpg")
        self.assertEqual(img, "imageid")

        self.assertEqual(mock.call_count, 3)
        self.assertEqual(
            mock.request_history[0].json(), {"hash": "29a26cadf018c8bd5eb4e2eca4d58683"}
        )
        self.assertEqual(mock.request_history[2].json(), {"status": "checked"})

    def test_create_image_exists(self, mock):
        mock.post("/api/v1/image/", status_code=400, json={"id": "imageid"})
        mock.get("/api/v1/image/imageid/", json={"id": "imageid"})
        mock.patch("/api/v1/image/imageid/", status_code=200)

        fi = FileImport("processid")
        img = fi.create_image(SAMPLES / "200x200.jpg")
        self.assertEqual(img, "imageid")

        self.assertEqual(mock.call_count, 3)
        self.assertEqual(
            mock.request_history[0].json(), {"hash": "29a26cadf018c8bd5eb4e2eca4d58683"}
        )
        self.assertEqual(mock.request_history[2].json(), {"status": "checked"})

    def test_create_image_error(self, mock):
        """
        Test that create_image only handles HTTP 400 errors
        """
        mock.post("/api/v1/image/", status_code=500)

        fi = FileImport("processid")
        with self.assertRaises(ErrorResponse):
            fi.create_image(SAMPLES / "200x200.jpg")

        self.assertEqual(mock.call_count, 1)
        self.assertEqual(
            mock.request_history[0].json(), {"hash": "29a26cadf018c8bd5eb4e2eca4d58683"}
        )

    def test_upload_retry(self, mock):
        mock.put(
            "http://s3/put",
            [{"exc": ConnectionError}, {"status_code": 500}, {"status_code": 201}],
        )

        fi = FileImport("processid")
        fi.upload_image(SAMPLES / "200x200.jpg", "http://s3/put")

        self.assertEqual(mock.call_count, 3)

    def test_upload_retry_exhausted(self, mock):
        mock.put("http://s3/put", exc=ConnectionError)

        fi = FileImport("processid")
        with self.assertRaises(ConnectionError):
            fi.upload_image(SAMPLES / "200x200.jpg", "http://s3/put")

        self.assertEqual(mock.call_count, 3)

    @patch("arkindex_tasks.import_files.base.datetime")
    @patch(
        "arkindex_tasks.import_files.base.WORKER_RUN_ID",
        "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
    )
    def test_create_top_level_folder(self, mock, datetime_mock):
        mock.post(
            "/api/v1/elements/create/",
            json={"id": "folderid", "type": "folder", "name": "Folder"},
        )
        datetime_mock.now.return_value = datetime(2020, 2, 2, 20, 2)
        fi = FileImport("processid")
        fi.process = {
            "id": "processid",
            "mode": "files",
            "corpus": "corpusid",
            "element": None,
            "folder_type": "folder",
            "element_type": "page",
            "files": ["file1", "file2"],
        }
        fi.create_top_level_folder(
            [
                {"id": "file1", "name": "file.jpg", "content_type": "image/jpeg"},
                {"id": "file2", "name": "file2.jpg", "content_type": "image/jpeg"},
            ]
        )
        self.assertEqual(mock.call_count, 1)
        self.assertEqual(
            mock.request_history[0].json(),
            {
                "corpus": "corpusid",
                "type": "folder",
                "name": "Import images 2020-02-02 20:02",
                "worker_run_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            },
        )
        self.assertEqual(
            fi.process["element"],
            {"id": "folderid", "type": "folder", "name": "Folder"},
        )

    @patch(
        "arkindex_tasks.import_files.base.WORKER_RUN_ID",
        "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
    )
    def test_create_top_level_folder_datafile_name(self, mock):
        mock.post(
            "/api/v1/elements/create/",
            json={"id": "folderid", "type": "folder", "name": "Folder"},
        )
        fi = FileImport("processid")
        fi.process = {
            "id": "processid",
            "mode": "files",
            "corpus": "corpusid",
            "element": None,
            "folder_type": "folder",
            "element_type": "page",
            "files": ["file1"],
        }
        fi.create_top_level_folder(
            [{"id": "file1", "name": "file.jpg", "content_type": "image/jpeg"}]
        )
        self.assertEqual(mock.call_count, 1)
        self.assertEqual(
            mock.request_history[0].json(),
            {
                "corpus": "corpusid",
                "type": "folder",
                "name": "file.jpg",
                "worker_run_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            },
        )
        self.assertEqual(
            fi.process["element"],
            {"id": "folderid", "type": "folder", "name": "Folder"},
        )

    def test_create_top_level_folder_ignore_pdf(self, mock):
        fi = FileImport("processid")
        fi.process = {
            "id": "processid",
            "mode": "files",
            "corpus": "corpusid",
            "element": None,
            "folder_type": "folder",
            "element_type": "page",
            "files": ["file1"],
        }
        fi.create_top_level_folder(
            [{"id": "file1", "name": "file.pdf", "content_type": "application/pdf"}]
        )
        self.assertEqual(mock.call_count, 0)
        self.assertIsNone(fi.process["element"])

    def test_create_top_level_folder_ignore_element(self, mock):
        fi = FileImport("processid")
        fi.process = {
            "id": "processid",
            "mode": "files",
            "corpus": "corpusid",
            "element": {"id": "folderid", "name": "My folder"},
            "folder_type": "folder",
            "element_type": "page",
            "files": ["file1"],
        }
        fi.create_top_level_folder([{"id": "file1", "name": "file.jpg"}])
        self.assertEqual(mock.call_count, 0)

    def test_create_element_top_level_error(self, mock):
        fi = FileImport("processid")
        with self.assertRaisesRegex(
            AssertionError, "Cannot create an image as a top-level element."
        ):
            fi.create_element(image_path="some_image", name="some_name", parent_id=None)

    @patch(
        "arkindex_tasks.import_files.base.WORKER_RUN_ID",
        "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
    )
    def test_save_elements(self, mock):
        mock.post(
            "/api/v1/image/",
            json={
                "id": "imageid",
                "s3_put_url": "http://s3/put",
                "s3_url": "http://s3/200x200.jpg",
            },
        )
        mock.put("http://s3/put", status_code=201)
        mock.patch("/api/v1/image/imageid/", status_code=200)
        mock.post("/api/v1/elements/create/", json={"id": "pageid", "type": "page"})

        fi = FileImport("processid")
        fi.process = {
            "id": "processid",
            "corpus": "corpusid",
            "element": {"id": "folderid", "type": "potato"},
            "folder_type": "folder",
            "element_type": "page",
        }

        datafile = {
            "id": "file1",
            "name": "200x200.jpg",
            "s3_url": "http://s3/200x200.jpg",
        }
        pages = fi.save_elements(
            datafile, [SAMPLES / "200x200.jpg"], parent_id="folderid"
        )

        self.assertListEqual(
            pages,
            [{"id": "pageid", "type": "page"}],
        )

        self.assertEqual(mock.call_count, 4)
        self.assertEqual(
            mock.request_history[0].json(), {"hash": "29a26cadf018c8bd5eb4e2eca4d58683"}
        )
        self.assertEqual(mock.request_history[2].json(), {"status": "checked"})
        self.assertEqual(
            mock.request_history[3].json(),
            {
                "corpus": "corpusid",
                "type": "page",
                "name": "200x200.jpg",
                "image": "imageid",
                "parent": "folderid",
                "worker_run_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            },
        )

    @patch(
        "arkindex_tasks.import_files.base.WORKER_RUN_ID",
        "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
    )
    def test_save_elements_top_level(self, mock):
        """
        Assert that save_elements creates folders even when a datafile has a single image
        if the process runs on a corpus, without a parent element
        """
        mock.post(
            "/api/v1/image/",
            json={
                "id": "imageid",
                "s3_put_url": "http://s3/put",
                "s3_url": "http://s3/200x200.jpg",
            },
        )
        mock.put("http://s3/put", status_code=201)
        mock.patch("/api/v1/image/imageid/", status_code=200)
        mock.post(
            "/api/v1/elements/create/",
            [
                {"json": {"id": "folderid", "type": "folder"}},
                {"json": {"id": "pageid", "type": "page"}},
            ],
        )

        fi = FileImport("processid")
        fi.process = {
            "id": "processid",
            "corpus": "corpusid",
            "element": None,
            "folder_type": "folder",
            "element_type": "page",
        }

        datafile = {
            "id": "file1",
            "name": "200x200.jpg",
            "s3_url": "http://s3/200x200.jpg",
        }
        elements = fi.save_elements(datafile, [SAMPLES / "200x200.jpg"], parent_id=None)

        self.assertListEqual(
            elements,
            [{"id": "folderid", "type": "folder"}, {"id": "pageid", "type": "page"}],
        )

        self.assertEqual(mock.call_count, 5)
        self.assertEqual(
            mock.request_history[0].json(),
            {
                "corpus": "corpusid",
                "type": "folder",
                "name": "200x200.jpg",
                "parent": None,
                "worker_run_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            },
        )
        self.assertEqual(
            mock.request_history[1].json(), {"hash": "29a26cadf018c8bd5eb4e2eca4d58683"}
        )
        self.assertEqual(mock.request_history[3].json(), {"status": "checked"})
        self.assertEqual(
            mock.request_history[4].json(),
            {
                "corpus": "corpusid",
                "type": "page",
                "name": "1",
                "image": "imageid",
                "parent": "folderid",
                "worker_run_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            },
        )

    @patch(
        "arkindex_tasks.import_files.base.WORKER_RUN_ID",
        "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
    )
    def test_run_image(self, mock):
        # Process info
        mock.get(
            "/api/v1/process/processid/",
            json={
                "id": "processid",
                "corpus": "corpusid",
                "mode": "files",
                "element": {"id": "folderid", "name": "Untitled", "type": "potato"},
                "files": ["file1"],
                "folder_type": "potato",
                "element_type": "page",
            },
        )
        # DataFile info
        mock.get(
            "/api/v1/process/file/file1/",
            json={
                "id": "file1",
                "name": "600x600.jpg",
                "s3_url": "http://s3/600x600.jpg",
                "content_type": "image/jpeg",
            },
        )
        # DataFile on S3
        mock.get("http://s3/600x600.jpg", body=open(SAMPLES / "600x600.jpg", "rb"))
        # Image creation
        mock.post(
            "/api/v1/image/",
            json={
                "id": "imageid",
                "s3_put_url": "http://s3/put",
                "s3_url": "http://s3/600x600.jpg",
            },
        )
        mock.put("http://s3/put", status_code=201)
        mock.patch("/api/v1/image/imageid/", status_code=200)
        # Page creation
        mock.post("/api/v1/elements/create/", json={"id": "pageid", "type": "page"})
        # DataFile deletion
        mock.delete("/api/v1/process/file/file1/", status_code=204)
        # Process update
        mock.patch("/api/v1/process/processid/", status_code=200)

        fi = FileImport("processid")
        fi.run()

        self.assertListEqual(
            [(req.method, req.url) for req in mock.request_history],
            [
                ("GET", "https://arkindex.teklia.com/api/v1/process/processid/"),
                ("GET", "https://arkindex.teklia.com/api/v1/process/file/file1/"),
                ("GET", "http://s3/600x600.jpg"),
                ("POST", "https://arkindex.teklia.com/api/v1/image/"),
                ("PUT", "http://s3/put"),
                ("PATCH", "https://arkindex.teklia.com/api/v1/image/imageid/"),
                ("POST", "https://arkindex.teklia.com/api/v1/elements/create/"),
                ("PATCH", "https://arkindex.teklia.com/api/v1/process/processid/"),
                ("DELETE", "https://arkindex.teklia.com/api/v1/process/file/file1/"),
            ],
        )

        with (get_working_dir() / "elements.json").open() as f:
            elements = json.load(f)
        self.assertListEqual(
            elements,
            [{"id": "folderid", "type": "potato"}, {"id": "pageid", "type": "page"}],
        )

    @patch(
        "arkindex_tasks.import_files.base.WORKER_RUN_ID",
        "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
    )
    @patch(
        "arkindex_tasks.import_files.pdf.WORKER_RUN_ID",
        "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
    )
    def test_run_pdf(self, mock):
        # Process info
        self.maxDiff = None
        mock.get(
            "/api/v1/process/processid/",
            json={
                "id": "processid",
                "corpus": "corpusid",
                "mode": "files",
                "element": {"id": "folderid", "name": "Untitled", "type": "book"},
                "files": ["file1", "file2"],
                "element_type": "page",
                "folder_type": "book",
            },
        )
        # DataFile info
        mock.get(
            "/api/v1/process/file/file1/",
            json={
                "id": "file1",
                "name": "file.pdf",
                "content_type": "application/pdf",
                "s3_url": "http://s3/file.pdf",
            },
        )
        mock.get(
            "/api/v1/process/file/file2/",
            json={
                "id": "file2",
                "name": "file2.pdf",
                "content_type": "application/pdf",
                "s3_url": "http://s3/file2.pdf",
            },
        )
        # DataFile on S3
        mock.get("http://s3/file.pdf", body=open(SAMPLES / "file.pdf", "rb"))
        mock.get("http://s3/file2.pdf", body=open(SAMPLES / "file2.pdf", "rb"))
        # Image creation
        mock.post(
            "/api/v1/image/",
            json={
                "id": "imageid",
                "s3_put_url": "http://s3/put",
                "s3_url": "http://s3/image.jpg",
            },
        )
        mock.put("http://s3/put", status_code=201)
        mock.patch("/api/v1/image/imageid/", status_code=200)
        # Page creation
        mock.post(
            "/api/v1/elements/create/",
            [
                {"json": {"id": "box1", "type": "book"}},
                {
                    "json": {
                        "id": "page1",
                        "type": "page",
                        "zone": {
                            "image": {
                                "width": 1200,
                                "height": 3000,
                            }
                        },
                    }
                },
                {
                    "json": {
                        "id": "page2",
                        "type": "page",
                        "zone": {
                            "image": {
                                "width": 1200,
                                "height": 3000,
                            }
                        },
                    }
                },
                {"json": {"id": "box2", "type": "book"}},
                {
                    "json": {
                        "id": "page1-1",
                        "type": "page",
                        "zone": {
                            "image": {
                                "width": 1200,
                                "height": 3000,
                            }
                        },
                    }
                },
                {
                    "json": {
                        "id": "page2-1",
                        "type": "page",
                        "zone": {
                            "image": {
                                "width": 1200,
                                "height": 3000,
                            }
                        },
                    }
                },
            ],
        )
        # Process update
        mock.patch("/api/v1/process/processid/", status_code=200)
        # DataFile deletion
        mock.delete("/api/v1/process/file/file1/", status_code=204)
        mock.delete("/api/v1/process/file/file2/", status_code=204)

        mock.post(
            "/api/v1/element/page1/transcriptions/bulk/",
        )
        mock.post(
            "/api/v1/element/page2/transcriptions/bulk/",
        )
        mock.post(
            "/api/v1/element/page1-1/transcriptions/bulk/",
        )
        mock.post(
            "/api/v1/element/page2-1/transcriptions/bulk/",
        )

        fi = FileImport("processid")
        fi.run()

        self.assertListEqual(
            [(req.method, req.url) for req in mock.request_history],
            [
                ("GET", "https://arkindex.teklia.com/api/v1/process/processid/"),
                ("GET", "https://arkindex.teklia.com/api/v1/process/file/file1/"),
                ("GET", "http://s3/file.pdf"),
                ("GET", "https://arkindex.teklia.com/api/v1/process/file/file2/"),
                ("GET", "http://s3/file2.pdf"),
                ("POST", "https://arkindex.teklia.com/api/v1/elements/create/"),
                ("POST", "https://arkindex.teklia.com/api/v1/image/"),
                ("PUT", "http://s3/put"),
                ("PATCH", "https://arkindex.teklia.com/api/v1/image/imageid/"),
                ("POST", "https://arkindex.teklia.com/api/v1/elements/create/"),
                ("POST", "https://arkindex.teklia.com/api/v1/image/"),
                ("PUT", "http://s3/put"),
                ("PATCH", "https://arkindex.teklia.com/api/v1/image/imageid/"),
                ("POST", "https://arkindex.teklia.com/api/v1/elements/create/"),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/page1/transcriptions/bulk/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/page2/transcriptions/bulk/",
                ),
                ("POST", "https://arkindex.teklia.com/api/v1/elements/create/"),
                ("POST", "https://arkindex.teklia.com/api/v1/image/"),
                ("PUT", "http://s3/put"),
                ("PATCH", "https://arkindex.teklia.com/api/v1/image/imageid/"),
                ("POST", "https://arkindex.teklia.com/api/v1/elements/create/"),
                ("POST", "https://arkindex.teklia.com/api/v1/image/"),
                ("PUT", "http://s3/put"),
                ("PATCH", "https://arkindex.teklia.com/api/v1/image/imageid/"),
                ("POST", "https://arkindex.teklia.com/api/v1/elements/create/"),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/page1-1/transcriptions/bulk/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/page2-1/transcriptions/bulk/",
                ),
                ("PATCH", "https://arkindex.teklia.com/api/v1/process/processid/"),
                ("DELETE", "https://arkindex.teklia.com/api/v1/process/file/file2/"),
                ("DELETE", "https://arkindex.teklia.com/api/v1/process/file/file1/"),
            ],
        )

        with (get_working_dir() / "elements.json").open() as f:
            elements = json.load(f)
        self.assertListEqual(
            elements,
            [
                {"id": "folderid", "type": "book"},
                {"id": "box1", "type": "book"},
                {
                    "id": "page1",
                    "type": "page",
                    "zone": {
                        "image": {
                            "width": 1200,
                            "height": 3000,
                        }
                    },
                },
                {
                    "id": "page2",
                    "type": "page",
                    "zone": {
                        "image": {
                            "width": 1200,
                            "height": 3000,
                        }
                    },
                },
                {"id": "box2", "type": "book"},
                {
                    "id": "page1-1",
                    "type": "page",
                    "zone": {
                        "image": {
                            "width": 1200,
                            "height": 3000,
                        }
                    },
                },
                {
                    "id": "page2-1",
                    "type": "page",
                    "zone": {
                        "image": {
                            "width": 1200,
                            "height": 3000,
                        }
                    },
                },
            ],
        )

    @patch(
        "arkindex_tasks.import_files.base.WORKER_RUN_ID",
        "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
    )
    @patch(
        "arkindex_tasks.import_files.pdf.WORKER_RUN_ID",
        "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
    )
    def test_run_pdf_and_image(self, mock):
        # Process info
        mock.get(
            "/api/v1/process/processid/",
            json={
                "id": "processid",
                "corpus": "corpusid",
                "mode": "files",
                "element": {"id": "folderid", "name": "Untitled", "type": "book"},
                "files": ["file1", "file2"],
                "element_type": "page",
                "folder_type": "book",
            },
        )
        # DataFiles info
        mock.get(
            "/api/v1/process/file/file1/",
            json={
                "id": "file1",
                "name": "file.pdf",
                "content_type": "application/pdf",
                "s3_url": "http://s3/file.pdf",
            },
        )
        mock.get(
            "/api/v1/process/file/file2/",
            json={
                "id": "file2",
                "name": "file.jpg",
                "content_type": "image/jpeg",
                "s3_url": "http://s3/600x600.jpg",
            },
        )
        # DataFiles on S3
        mock.get("http://s3/file.pdf", body=open(SAMPLES / "file.pdf", "rb"))
        mock.get("http://s3/600x600.jpg", body=open(SAMPLES / "600x600.jpg", "rb"))
        # Images creation
        mock.post(
            "/api/v1/image/",
            json={
                "id": "imageid",
                "s3_put_url": "http://s3/put",
                "s3_url": "http://s3/image.jpg",
            },
        )
        mock.put("http://s3/put", status_code=201)
        mock.patch("/api/v1/image/imageid/", status_code=200)
        mock.put("http://s3/put", status_code=201)
        mock.patch("/api/v1/image/imageid2/", status_code=200)
        # Elements creation
        mock.post(
            "/api/v1/elements/create/",
            [
                {
                    "json": {
                        "id": "pageid",
                        "type": "page",
                        "zone": {
                            "image": {
                                "width": 600,
                                "height": 600,
                            }
                        },
                    }
                },
                {"json": {"id": "box", "type": "book"}},
                {
                    "json": {
                        "id": "page1",
                        "type": "page",
                        "zone": {
                            "image": {
                                "width": 1200,
                                "height": 3000,
                            }
                        },
                    }
                },
                {
                    "json": {
                        "id": "page2",
                        "type": "page",
                        "zone": {
                            "image": {
                                "width": 1200,
                                "height": 3000,
                            }
                        },
                    }
                },
            ],
        )
        # Process update
        mock.patch("/api/v1/process/processid/", status_code=200)
        # DataFile deletion
        mock.delete("/api/v1/process/file/file1/", status_code=204)
        mock.delete("/api/v1/process/file/file2/", status_code=204)
        # PDF transcriptions creation
        mock.post(
            "/api/v1/element/page1/transcriptions/bulk/",
        )
        mock.post(
            "/api/v1/element/page2/transcriptions/bulk/",
        )

        fi = FileImport("processid")
        fi.run()

        self.assertListEqual(
            [(req.method, req.url) for req in mock.request_history],
            [
                ("GET", "https://arkindex.teklia.com/api/v1/process/processid/"),
                ("GET", "https://arkindex.teklia.com/api/v1/process/file/file1/"),
                ("GET", "http://s3/file.pdf"),
                ("GET", "https://arkindex.teklia.com/api/v1/process/file/file2/"),
                ("GET", "http://s3/600x600.jpg"),
                ("POST", "https://arkindex.teklia.com/api/v1/image/"),
                ("PUT", "http://s3/put"),
                ("PATCH", "https://arkindex.teklia.com/api/v1/image/imageid/"),
                ("POST", "https://arkindex.teklia.com/api/v1/elements/create/"),
                ("POST", "https://arkindex.teklia.com/api/v1/elements/create/"),
                ("POST", "https://arkindex.teklia.com/api/v1/image/"),
                ("PUT", "http://s3/put"),
                ("PATCH", "https://arkindex.teklia.com/api/v1/image/imageid/"),
                ("POST", "https://arkindex.teklia.com/api/v1/elements/create/"),
                ("POST", "https://arkindex.teklia.com/api/v1/image/"),
                ("PUT", "http://s3/put"),
                ("PATCH", "https://arkindex.teklia.com/api/v1/image/imageid/"),
                ("POST", "https://arkindex.teklia.com/api/v1/elements/create/"),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/page1/transcriptions/bulk/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/page2/transcriptions/bulk/",
                ),
                ("PATCH", "https://arkindex.teklia.com/api/v1/process/processid/"),
                ("DELETE", "https://arkindex.teklia.com/api/v1/process/file/file2/"),
                ("DELETE", "https://arkindex.teklia.com/api/v1/process/file/file1/"),
            ],
        )

        with (get_working_dir() / "elements.json").open() as f:
            elements = json.load(f)
        self.assertListEqual(
            elements,
            [
                {"id": "folderid", "type": "book"},
                {
                    "id": "pageid",
                    "type": "page",
                    "zone": {
                        "image": {
                            "width": 600,
                            "height": 600,
                        }
                    },
                },
                {"id": "box", "type": "book"},
                {
                    "id": "page1",
                    "type": "page",
                    "zone": {
                        "image": {
                            "width": 1200,
                            "height": 3000,
                        }
                    },
                },
                {
                    "id": "page2",
                    "type": "page",
                    "zone": {
                        "image": {
                            "width": 1200,
                            "height": 3000,
                        }
                    },
                },
            ],
        )

    @patch(
        "arkindex_tasks.import_files.base.WORKER_RUN_ID",
        "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
    )
    def test_run_top_level(self, mock):
        # Process info
        mock.get(
            "/api/v1/process/processid/",
            json={
                "id": "processid",
                "corpus": "corpusid",
                "mode": "files",
                "element": None,
                "files": ["file1"],
                "folder_type": "folder",
                "element_type": "page",
            },
        )
        # DataFile info
        mock.get(
            "/api/v1/process/file/file1/",
            json={
                "id": "file1",
                "name": "600x600.jpg",
                "s3_url": "http://s3/600x600.jpg",
                "content_type": "image/jpeg",
            },
        )
        # DataFile on S3
        mock.get("http://s3/600x600.jpg", body=open(SAMPLES / "600x600.jpg", "rb"))
        # Image creation
        mock.post(
            "/api/v1/image/",
            json={
                "id": "imageid",
                "s3_put_url": "http://s3/put",
                "s3_url": "http://s3/600x600.jpg",
            },
        )
        mock.put("http://s3/put", status_code=201)
        mock.patch("/api/v1/image/imageid/", status_code=200)
        # Folder and page creation
        mock.post(
            "/api/v1/elements/create/",
            [
                {"json": {"id": "folderid", "type": "folder", "name": "Folder"}},
                {"json": {"id": "pageid", "type": "page"}},
            ],
        )
        # Process update
        mock.patch("/api/v1/process/processid/", status_code=200)
        # DataFile deletion
        mock.delete("/api/v1/process/file/file1/", status_code=204)

        fi = FileImport("processid")
        fi.run()

        self.assertListEqual(
            [(req.method, req.url) for req in mock.request_history],
            [
                ("GET", "https://arkindex.teklia.com/api/v1/process/processid/"),
                ("GET", "https://arkindex.teklia.com/api/v1/process/file/file1/"),
                ("GET", "http://s3/600x600.jpg"),
                ("POST", "https://arkindex.teklia.com/api/v1/elements/create/"),
                ("POST", "https://arkindex.teklia.com/api/v1/image/"),
                ("PUT", "http://s3/put"),
                ("PATCH", "https://arkindex.teklia.com/api/v1/image/imageid/"),
                ("POST", "https://arkindex.teklia.com/api/v1/elements/create/"),
                ("PATCH", "https://arkindex.teklia.com/api/v1/process/processid/"),
                ("DELETE", "https://arkindex.teklia.com/api/v1/process/file/file1/"),
            ],
        )

        with (get_working_dir() / "elements.json").open() as f:
            elements = json.load(f)
        self.assertListEqual(
            elements,
            [{"id": "folderid", "type": "folder"}, {"id": "pageid", "type": "page"}],
        )

    @patch(
        "arkindex_tasks.import_files.base.WORKER_RUN_ID",
        "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
    )
    def test_run_sorted(self, mock):
        """
        Ensure that files are sorted by name using natsort
        """
        numbers = [1, 2, 10, 100, 234]
        self.assertListEqual(sorted(numbers), [1, 2, 10, 100, 234])
        self.assertListEqual(sorted(map(str, numbers)), ["1", "10", "100", "2", "234"])

        # Process info
        mock.get(
            "/api/v1/process/processid/",
            json={
                "id": "processid",
                "corpus": "corpusid",
                "mode": "files",
                "element": {"id": "folderid", "name": "Untitled", "type": "folder"},
                "files": list(map(str, numbers)),
                "folder_type": "folder",
                "element_type": "page",
            },
        )

        # DataFile info and file on S3
        for i in numbers:
            mock.get(
                f"/api/v1/process/file/{i}/",
                json={
                    "id": str(i),
                    "name": f"file_{i}.jpg",
                    "s3_url": f"http://s3/file_{i}.jpg",
                    "content_type": "image/jpeg",
                },
            )
            mock.get(
                f"http://s3/file_{i}.jpg", body=open(SAMPLES / "600x600.jpg", "rb")
            )

        # Image creation
        mock.post(
            "/api/v1/image/",
            json={
                "id": "imageid",
                "s3_put_url": "http://s3/put",
                "s3_url": "http://s3/600x600.jpg",
            },
        )
        mock.put("http://s3/put", status_code=201)
        mock.patch("/api/v1/image/imageid/", status_code=200)

        # Page creation
        mock.post("/api/v1/elements/create/", json={"id": "pageid", "type": "page"})

        # Process element update
        mock.patch("/api/v1/process/processid/", status_code=200)

        # DataFile deletion
        for i in numbers:
            mock.delete(f"/api/v1/process/file/{i}/", status_code=204)

        fi = FileImport("processid")
        fi.run()

        expected_queries = [
            ("GET", "https://arkindex.teklia.com/api/v1/process/processid/"),
            ("GET", "https://arkindex.teklia.com/api/v1/process/file/1/"),
            ("GET", "http://s3/file_1.jpg"),
            ("GET", "https://arkindex.teklia.com/api/v1/process/file/2/"),
            ("GET", "http://s3/file_2.jpg"),
            ("GET", "https://arkindex.teklia.com/api/v1/process/file/10/"),
            ("GET", "http://s3/file_10.jpg"),
            ("GET", "https://arkindex.teklia.com/api/v1/process/file/100/"),
            ("GET", "http://s3/file_100.jpg"),
            ("GET", "https://arkindex.teklia.com/api/v1/process/file/234/"),
            ("GET", "http://s3/file_234.jpg"),
        ]
        expected_queries += [
            ("POST", "https://arkindex.teklia.com/api/v1/image/"),
            ("PUT", "http://s3/put"),
            ("PATCH", "https://arkindex.teklia.com/api/v1/image/imageid/"),
            ("POST", "https://arkindex.teklia.com/api/v1/elements/create/"),
        ] * 5
        expected_queries += [
            ("PATCH", "https://arkindex.teklia.com/api/v1/process/processid/"),
            ("DELETE", "https://arkindex.teklia.com/api/v1/process/file/1/"),
            ("DELETE", "https://arkindex.teklia.com/api/v1/process/file/2/"),
            ("DELETE", "https://arkindex.teklia.com/api/v1/process/file/10/"),
            ("DELETE", "https://arkindex.teklia.com/api/v1/process/file/100/"),
            ("DELETE", "https://arkindex.teklia.com/api/v1/process/file/234/"),
        ]

        self.assertListEqual(
            [(req.method, req.url) for req in mock.request_history], expected_queries
        )

        with (get_working_dir() / "elements.json").open() as f:
            elements = json.load(f)
        self.assertListEqual(
            elements,
            [
                {"id": "folderid", "type": "folder"},
                {"id": "pageid", "type": "page"},
                {"id": "pageid", "type": "page"},
                {"id": "pageid", "type": "page"},
                {"id": "pageid", "type": "page"},
                {"id": "pageid", "type": "page"},
            ],
        )

    @patch(
        "arkindex_tasks.import_files.base.WORKER_RUN_ID",
        "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
    )
    @patch("arkindex_tasks.import_files.base.TranskribusImporter")
    @patch("arkindex_tasks.import_files.base.logger.error")
    def test_run_zip(self, mock, logger_mock, importer_mock):
        # Process info
        mock.get(
            "/api/v1/process/processid/",
            json={
                "id": "processid",
                "corpus": "corpusid",
                "mode": "files",
                "element": {"id": "folderid", "name": "Untitled", "type": "book"},
                "files": ["file1", "file2", "file3", "file4"],
                "element_type": "page",
                "folder_type": "book",
            },
        )
        # DataFile info
        mock.get(
            "/api/v1/process/file/file1/",
            json={
                "id": "file1",
                "name": "file1.zip",
                "content_type": "application/zip",
                "s3_url": "http://s3/file1.zip",
            },
        )
        mock.get(
            "/api/v1/process/file/file2/",
            json={
                "id": "file2",
                "name": "file2.zip",
                "content_type": "application/zip",
                "s3_url": "http://s3/file2.zip",
            },
        )
        mock.get(
            "/api/v1/process/file/file3/",
            json={
                "id": "file3",
                "name": "file3.zip",
                # Use the Windows MIME type
                "content_type": "application/x-zip-compressed",
                "s3_url": "http://s3/file3.zip",
            },
        )
        mock.get(
            "/api/v1/process/file/file4/",
            json={
                "id": "file4",
                "name": "file4.zip",
                "content_type": "application/zip",
                "s3_url": "http://s3/file4.zip",
            },
        )

        # Create an archive that looks like a Transkribus export
        transkribus_zip = TemporaryFile()
        with ZipFile(transkribus_zip, "a") as archive:
            archive.writestr("1234/document_title/mets.xml", "lol")
        transkribus_zip.flush()
        transkribus_zip.seek(0)
        mock.get("http://s3/file1.zip", body=transkribus_zip)

        # An archive that isn't a Transkribus export
        dummy_zip = TemporaryFile()
        with ZipFile(dummy_zip, "a") as archive:
            archive.writestr("lol.txt", "lol")
        dummy_zip.flush()
        dummy_zip.seek(0)
        mock.get("http://s3/file2.zip", body=dummy_zip)

        # Windows MIME type test: we just use a dummy archive.
        # We cannot re-use the dummy_zip defined above since it will be consumed by the mocked request
        windows_zip = TemporaryFile()
        with ZipFile(windows_zip, "a") as archive:
            archive.writestr("lol.txt", "lol")
        windows_zip.flush()
        windows_zip.seek(0)
        mock.get("http://s3/file3.zip", body=windows_zip)

        # Not actually a zip file
        mock.get("http://s3/file4.zip", body=open(SAMPLES / "file2.pdf", "rb"))

        mock.post(
            "/api/v1/elements/create/",
            [
                {"json": {"id": "box1", "type": "book"}},
                {"json": {"id": "box2", "type": "book"}},
            ],
        )
        # Process update
        mock.patch("/api/v1/process/processid/", status_code=200)
        # Only file1 gets deleted, since the other files will not be imported successfully
        mock.delete("/api/v1/process/file/file1/", status_code=204)

        importer_mock.run.return_value = [{"id": "imported", "type": "thing"}]

        fi = FileImport("processid")
        fi.run()

        # file1.zip triggers a Transkribus import
        importer_mock.assert_called_once()

        # The only positional argument is a path to the temporary file
        # that the file import uses to download the file.
        self.assertEqual(len(importer_mock.call_args.args), 1)
        # To check that this is the Transkribus export, we open the archive and look for the XML file we put in there
        with ZipFile(importer_mock.call_args.args[0]) as archive:
            self.assertSequenceEqual(
                archive.namelist(), ["1234/document_title/mets.xml"]
            )

        self.assertDictEqual(
            importer_mock.call_args.kwargs,
            {
                "corpus": "corpusid",
                "parent_id": "folderid",
                "folder_type": "book",
                "element_type": "page",
            },
        )

        self.assertListEqual(
            logger_mock.call_args_list,
            [
                # Dummy ZIP file
                call(
                    "Could not import file lol.txt from archive file2.zip: NotImplementedError: "
                ),
                call(
                    "Could not import file file2.zip: AssertionError: Not all files from archive file2.zip have been imported."
                ),
                # Dummy ZIP file with Windows MIME type
                call(
                    "Could not import file lol.txt from archive file3.zip: NotImplementedError: "
                ),
                call(
                    "Could not import file file3.zip: AssertionError: Not all files from archive file3.zip have been imported."
                ),
                # Not a ZIP file
                call("File file4.zip does not appear to be a valid ZIP archive."),
                call(
                    "Could not import file file4.zip: BadZipFile: File is not a zip file"
                ),
            ],
        )

        self.assertListEqual(
            [(req.method, req.url) for req in mock.request_history],
            [
                ("GET", "https://arkindex.teklia.com/api/v1/process/processid/"),
                ("GET", "https://arkindex.teklia.com/api/v1/process/file/file1/"),
                ("GET", "http://s3/file1.zip"),
                ("GET", "https://arkindex.teklia.com/api/v1/process/file/file2/"),
                ("GET", "http://s3/file2.zip"),
                ("GET", "https://arkindex.teklia.com/api/v1/process/file/file3/"),
                ("GET", "http://s3/file3.zip"),
                ("GET", "https://arkindex.teklia.com/api/v1/process/file/file4/"),
                ("GET", "http://s3/file4.zip"),
                ("POST", "https://arkindex.teklia.com/api/v1/elements/create/"),
                ("POST", "https://arkindex.teklia.com/api/v1/elements/create/"),
                ("POST", "https://arkindex.teklia.com/api/v1/elements/create/"),
                ("PATCH", "https://arkindex.teklia.com/api/v1/process/processid/"),
                ("DELETE", "https://arkindex.teklia.com/api/v1/process/file/file1/"),
            ],
        )

        with (get_working_dir() / "elements.json").open() as f:
            elements = json.load(f)
        self.assertListEqual(
            elements,
            [
                {"id": "folderid", "type": "book"},
            ],
        )

    @patch(
        "arkindex_tasks.import_files.base.WORKER_RUN_ID",
        "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
    )
    @patch("arkindex_tasks.import_files.base.logger.error")
    def test_run_tar(self, mock, logger_mock):
        # Prepare all of the archives to test
        cases = [
            ("bzip", "application/x-bzip", "w:bz2"),
            ("bzip2", "application/x-bzip2", "w:bz2"),
            ("gz", "application/gzip", "w:gz"),
            ("lzma", "application/x-lzma", "w:xz"),
            ("tar", "application/x-tar", "w:"),
            ("tar-gz", "application/x-gtar", "w:gz"),
            ("xz", "application/x-xz", "w:xz"),
        ]

        # tarfile does not support an equivalent of ZipFile.writestr, so we need an actual file to add to archives.
        with NamedTemporaryFile() as text_file:
            text_file.write(b"lol")
            text_file.flush()

            for file_id, content_type, mode in cases:
                archive_file = TemporaryFile()
                with tarfile.open(fileobj=archive_file, mode=mode) as archive:
                    archive.add(text_file.name, arcname="lol.txt")
                archive_file.flush()
                archive_file.seek(0)
                mock.get(f"http://s3/{file_id}", body=archive_file)

            # .tar.zst is handled separately, since tarfile doesn't support it
            with TemporaryFile() as archive_file:
                with tarfile.open(fileobj=archive_file, mode="w:") as archive:
                    archive.add(text_file.name, arcname="lol.txt")
                archive_file.flush()
                archive_file.seek(0)

                zst_archive = TemporaryFile()
                ZstdCompressor().copy_stream(archive_file, zst_archive)
                zst_archive.seek(0)
                mock.get("http://s3/zst", body=zst_archive)

            cases.append(("zst", "application/zstd", None))

        # Build RetrieveDataFile mocks
        for file_id, content_type, _ in cases:
            mock.get(
                f"/api/v1/process/file/{file_id}/",
                json={
                    "id": file_id,
                    "name": file_id,
                    "content_type": content_type,
                    "s3_url": f"http://s3/{file_id}",
                },
            )

        # Process info
        mock.get(
            "/api/v1/process/processid/",
            json={
                "id": "processid",
                "corpus": "corpusid",
                "mode": "files",
                "element": {"id": "folderid", "name": "Untitled", "type": "book"},
                "files": [file_id for file_id, _, _ in cases],
                "element_type": "page",
                "folder_type": "book",
            },
        )

        # Mock all folder creations
        mock.post(
            "/api/v1/elements/create/",
            json={"id": "newfolderid", "type": "book"},
        )
        # Mock PartialUpdateProcess to allow setting Process.element
        mock.patch("/api/v1/process/processid/", status_code=200)

        fi = FileImport("processid")
        fi.run()

        expected_logs = []
        expected_requests = [
            ("GET", "https://arkindex.teklia.com/api/v1/process/processid/"),
        ]

        # For each archive, one call to RetrieveDataFile and one to the S3 URL
        for file_id, _, _ in cases:
            expected_logs.extend(
                [
                    call(
                        f"Could not import file lol.txt from archive {file_id}: NotImplementedError: "
                    ),
                    call(
                        f"Could not import file {file_id}: AssertionError: Not all files from archive {file_id} have been imported."
                    ),
                ]
            )
            expected_requests.extend(
                [
                    (
                        "GET",
                        f"https://arkindex.teklia.com/api/v1/process/file/{file_id}/",
                    ),
                    ("GET", f"http://s3/{file_id}"),
                ]
            )

        # One folder creation per archive, only after all the files are downloaded
        expected_requests.extend(
            [
                ("POST", "https://arkindex.teklia.com/api/v1/elements/create/")
                for _ in cases
            ]
        )

        expected_requests.append(
            ("PATCH", "https://arkindex.teklia.com/api/v1/process/processid/")
        )

        self.assertListEqual(
            logger_mock.call_args_list,
            expected_logs,
        )
        self.assertListEqual(
            [(req.method, req.url) for req in mock.request_history],
            expected_requests,
        )

    @patch(
        "arkindex_tasks.import_files.base.WORKER_RUN_ID",
        "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
    )
    @patch("arkindex_tasks.import_files.base.logger.error")
    def test_zipception(self, mock, logger_mock):
        zip4 = TemporaryFile()
        with BytesIO() as zip1, BytesIO() as zip2, BytesIO() as zip3:
            with ZipFile(zip1, "w") as archive:
                archive.writestr("lol.txt", "lol")

            with ZipFile(zip2, "w") as archive:
                archive.writestr("lol.zip", zip1.getvalue())

            with ZipFile(zip3, "w") as archive:
                archive.writestr("lol.zip.zip", zip2.getvalue())

            with ZipFile(zip4, "w") as archive:
                archive.writestr("lol.zip.zip.zip", zip3.getvalue())

        zip4.flush()
        zip4.seek(0)
        mock.get("http://s3/file1.zip", body=zip4)

        # Process info
        mock.get(
            "/api/v1/process/processid/",
            json={
                "id": "processid",
                "corpus": "corpusid",
                "mode": "files",
                "element": {"id": "folderid", "name": "Untitled", "type": "book"},
                "files": ["file1"],
                "element_type": "page",
                "folder_type": "book",
            },
        )
        # DataFile info
        mock.get(
            "/api/v1/process/file/file1/",
            json={
                "id": "file1",
                "name": "file1.zip",
                "content_type": "application/zip",
                "s3_url": "http://s3/file1.zip",
            },
        )

        mock.post(
            "/api/v1/elements/create/",
            json={"id": "newfolderid", "type": "book"},
        )

        # Process update
        mock.patch("/api/v1/process/processid/", status_code=200)

        fi = FileImport("processid")
        fi.run()

        self.assertListEqual(
            logger_mock.call_args_list,
            [
                call(
                    "Could not import file lol.zip from archive lol.zip.zip: AssertionError: Maximum archive decompression depth reached"
                ),
                call(
                    "Could not import file lol.zip.zip from archive lol.zip.zip.zip: AssertionError: Not all files from archive lol.zip.zip have been imported."
                ),
                call(
                    "Could not import file lol.zip.zip.zip from archive file1.zip: AssertionError: Not all files from archive lol.zip.zip.zip have been imported."
                ),
                call(
                    "Could not import file file1.zip: AssertionError: Not all files from archive file1.zip have been imported."
                ),
            ],
        )

        self.assertListEqual(
            [(req.method, req.url) for req in mock.request_history],
            [
                ("GET", "https://arkindex.teklia.com/api/v1/process/processid/"),
                ("GET", "https://arkindex.teklia.com/api/v1/process/file/file1/"),
                ("GET", "http://s3/file1.zip"),
                ("POST", "https://arkindex.teklia.com/api/v1/elements/create/"),
                ("POST", "https://arkindex.teklia.com/api/v1/elements/create/"),
                ("POST", "https://arkindex.teklia.com/api/v1/elements/create/"),
                ("PATCH", "https://arkindex.teklia.com/api/v1/process/processid/"),
            ],
        )

        with (get_working_dir() / "elements.json").open() as f:
            elements = json.load(f)
        self.assertListEqual(
            elements,
            [
                {"id": "folderid", "type": "book"},
            ],
        )

    @patch("arkindex_tasks.import_files.base.IIIFParser")
    @patch(
        "arkindex_tasks.import_files.base.WORKER_RUN_ID",
        "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
    )
    def test_run_iiif(self, mock, parser_mock):
        # Process info
        mock.get(
            "/api/v1/process/processid/",
            json={
                "id": "processid",
                "corpus": "corpusid",
                "mode": "files",
                "element": {"id": "folderid", "name": "Untitled", "type": "potato"},
                "files": ["file1", "file2"],
                "folder_type": "potato",
                "element_type": "page",
            },
        )
        # DataFile info
        mock.get(
            "/api/v1/process/file/file1/",
            json={
                "id": "file1",
                "name": "manifest.json",
                "s3_url": "http://somewhere.com/IIIF/file1/manifest.json",
                "content_type": "application/json",
            },
        )
        mock.get(
            "/api/v1/process/file/file2/",
            json={
                "id": "file2",
                "name": "collection.json",
                "s3_url": "http://somewhere.com/IIIF/file2/collection.json",
                "content_type": "application/json",
            },
        )
        # DataFile on S3
        mock.get(
            "http://somewhere.com/IIIF/file2/collection.json",
            body=open(SAMPLES / "collection.json", "rb"),
        )
        mock.get(
            "http://somewhere.com/IIIF/file1/manifest.json",
            body=open(SAMPLES / "manifest.json", "rb"),
        )
        # DataFile deletion
        mock.delete("/api/v1/process/file/file1/", status_code=204)
        mock.delete("/api/v1/process/file/file2/", status_code=204)
        # Process update
        mock.patch("/api/v1/process/processid/", status_code=200)

        parser_mock.load.return_value.elements = [
            {"id": "elementid", "type": "element"}
        ]

        fi = FileImport("processid")
        fi.run()

        self.assertListEqual(
            [(req.method, req.url) for req in mock.request_history],
            [
                ("GET", "https://arkindex.teklia.com/api/v1/process/processid/"),
                ("GET", "https://arkindex.teklia.com/api/v1/process/file/file1/"),
                ("GET", "http://somewhere.com/IIIF/file1/manifest.json"),
                ("GET", "https://arkindex.teklia.com/api/v1/process/file/file2/"),
                ("GET", "http://somewhere.com/IIIF/file2/collection.json"),
                ("PATCH", "https://arkindex.teklia.com/api/v1/process/processid/"),
                ("DELETE", "https://arkindex.teklia.com/api/v1/process/file/file2/"),
                ("DELETE", "https://arkindex.teklia.com/api/v1/process/file/file1/"),
            ],
        )

        call1, call2 = parser_mock.load.call_args_list
        args, kwargs = call1
        self.assertEqual(len(args), 1)
        self.assertIsInstance(args[0], BufferedReader)
        self.assertDictEqual(
            kwargs,
            {
                "corpus_id": "corpusid",
                "folder_type": "potato",
                "element_type": "page",
                "parent_id": "folderid",
            },
        )
        args, kwargs = call2
        self.assertEqual(len(args), 1)
        self.assertIsInstance(args[0], BufferedReader)
        self.assertDictEqual(
            kwargs,
            {
                "corpus_id": "corpusid",
                "folder_type": "potato",
                "element_type": "page",
                "parent_id": "folderid",
            },
        )

        elements = get_working_dir() / "elements.json"
        self.assertTrue(elements.is_file())
        with elements.open() as f:
            self.assertListEqual(
                json.load(f),
                [
                    {"id": "folderid", "type": "potato"},
                    {"id": "elementid", "type": "element"},
                    {"id": "elementid", "type": "element"},
                ],
            )
