# -*- coding: utf-8 -*-
import tempfile
from pathlib import Path
from unittest import TestCase
from unittest.mock import patch

from PIL import Image

from arkindex_tasks.import_files.image import check_image

SAMPLES = Path(__file__).absolute().parent / "samples"
GLOBAL_SAMPLES = Path(__file__).absolute().parent.parent / "samples"


class TestImageImport(TestCase):
    def setUp(self):
        self.tmp_dir = tempfile.mkdtemp()

    def tearDown(self):
        Path(self.tmp_dir).rmdir()

    @patch("arkindex_tasks.import_files.image.tempfile")
    def test_check_valid_image(self, tempfile_mock):
        """
        Do not re-save valid images
        """
        test_file = {
            "id": "fileid",
            "name": "600x600.jpg",
            "local_path": SAMPLES / "600x600.jpg",
            "content_type": "image/jpeg",
        }
        tempfile_mock.mkdtemp.return_value = self.tmp_dir
        result = check_image(test_file)
        # The original file path is returned
        self.assertListEqual(result, [test_file["local_path"]])
        # There is nothing at what would have been the new file path
        empty_path = Path(self.tmp_dir) / "fileid"
        self.assertFalse(empty_path.is_file())
        img = Image.open(test_file["local_path"])
        self.assertEqual(img.format, "JPEG")

    def test_check_image_ioerror(self):
        with self.assertRaises(ValueError, msg="File is not a valid image"):
            check_image(
                {
                    "id": "fileid",
                    "name": "600x600.jpg",
                    "local_path": SAMPLES / "nope.nope",
                }
            )

    @patch("arkindex_tasks.import_files.image.tempfile")
    def test_check_valid_non_jpeg_image(self, tempfile_mock):
        test_file = {
            "id": "fileid",
            "name": "600x600.png",
            "local_path": SAMPLES / "600x600.png",
            "content_type": "image/x-png",
        }
        tempfile_mock.mkdtemp.return_value = self.tmp_dir
        result = check_image(test_file)
        self.assertListEqual(result, [test_file["local_path"]])
        img = Image.open(test_file["local_path"])
        self.assertEqual(img.format, "PNG")

    @patch("arkindex_tasks.import_files.image.tempfile")
    def test_check_image_convert_gif(self, tempfile_mock):
        tempfile_mock.mkdtemp.return_value = self.tmp_dir
        result = check_image(
            {
                "id": "fileid",
                "name": "600x600.gif",
                "local_path": SAMPLES / "600x600.gif",
                "content_type": "image/gif",
            }
        )
        expected_path = Path(self.tmp_dir) / "fileid.jpg"
        self.assertListEqual(result, [expected_path])
        self.assertTrue(expected_path.is_file())
        img = Image.open(expected_path)
        self.assertEqual(img.format, "JPEG")
        expected_path.unlink()

    @patch("arkindex_tasks.import_files.image.tempfile")
    def test_check_image_convert_tiff(self, tempfile_mock):
        tempfile_mock.mkdtemp.return_value = self.tmp_dir
        result = check_image(
            {
                "id": "fileid",
                "name": "image.tiff",
                "local_path": SAMPLES / "image.tiff",
                "content_type": "image/tiff",
            }
        )
        expected_path = Path(self.tmp_dir) / "fileid.jpg"
        self.assertListEqual(result, [expected_path])
        self.assertTrue(expected_path.is_file())
        img = Image.open(expected_path)
        self.assertEqual(img.format, "JPEG")
        expected_path.unlink()

    @patch("arkindex_tasks.import_files.image.tempfile")
    def test_check_image_exif_orientation(self, tempfile_mock):
        tempfile_mock.mkdtemp.return_value = self.tmp_dir
        original_path = GLOBAL_SAMPLES / "rotated.jpg"
        original_image = Image.open(original_path)

        # 0x0112 is the ID of the EXIF Orientation tag. 6 means "rotate 270° clockwise"
        # See https://www.digicamsoft.com/exif22/exif22/html/exif22_22.htm
        self.assertEqual(original_image.getexif().get(0x0112), 6)
        self.assertTupleEqual(original_image.size, (200, 100))

        result = check_image(
            {
                "id": "fileid",
                "name": "rotated.jpg",
                "local_path": original_path,
                "content_type": "image/jpeg",
            }
        )
        expected_path = Path(self.tmp_dir) / "fileid"
        self.assertListEqual(result, [expected_path])
        self.assertTrue(expected_path.is_file())
        img = Image.open(expected_path)
        self.assertEqual(img.format, "JPEG")

        # Orientation tag has been removed
        self.assertIsNone(img.getexif().get(0x0112))
        # Size has changed, implying the image was rotated
        self.assertEqual(img.size, (100, 200))

        expected_path.unlink()

    @patch("arkindex_tasks.import_files.image.tempfile")
    def test_check_image_orientation_topleft(self, tempfile_mock):
        """
        Ensure check_image ignores an EXIF Orientation of 1 as it is a no-op
        """
        tempfile_mock.mkdtemp.return_value = self.tmp_dir
        img = Image.open(GLOBAL_SAMPLES / "rotated.jpg")
        self.assertEqual(img.size, (200, 100))
        exif = img.getexif()
        exif[0x0112] = 1
        with tempfile.NamedTemporaryFile() as f:
            img.save(f, format="JPEG", exif=exif)
            result = check_image(
                {
                    "id": "fileid",
                    "name": "600x600.jpg",
                    "local_path": Path(f.name),
                    "content_type": "image/jpeg",
                }
            )
        # The original file path is returned
        self.assertListEqual(result, [Path(f.name)])
        # There is nothing at what would have been the new file path
        expected_path = Path(self.tmp_dir) / "fileid"
        self.assertFalse(expected_path.is_file())
        img = Image.open(GLOBAL_SAMPLES / "rotated.jpg")
        self.assertEqual(img.format, "JPEG")
        # Image size has not changed
        self.assertEqual(img.size, (200, 100))
