# -*- coding: utf-8 -*-
import json
from collections import namedtuple
from pathlib import Path
from unittest import TestCase
from unittest.mock import patch

import requests_mock
from lxml import etree

from arkindex_tasks.import_files.pagexml import PageXmlParser

EMPTY_XML = etree.fromstring(
    '<PcGts xmlns="http://schema.primaresearch.org/PAGE/gts/pagecontent/2013-07-15"><Page></Page></PcGts>'
)

SAMPLES = Path(__file__).absolute().parent.parent / "samples"

fake_region = namedtuple(
    "fake_region",
    ["id", "points", "text", "confidence"],
    defaults=["", None],
)

fake_image = {
    "id": "fake_image_id",
    "path": "/path/path/",
    "width": 120000,
    "height": 120000,
    "url": "http://example.com",
}
smaller_image = {
    "id": "bby_image_id",
    "path": "/path/path/path",
    "width": 300,
    "height": 160,
    "url": "http://example.com/image/",
}


class TestPageXmlParser(TestCase):
    def test_simplify_polygon(self):
        page_xml_parser = PageXmlParser(EMPTY_XML, fake_image)

        # Simple polygon
        points = [(10, 10), (10, 20), (20, 20), (10, 10)]
        region = fake_region("region_id", points)
        result = page_xml_parser.simplify_polygon(region)
        self.assertListEqual(result, points)

        # Polygon with 938 points
        points = json.loads((SAMPLES / "big_polygon.json").read_text())
        self.assertGreater(len(points), 512)

        # Polygon with 453 points
        expected = json.loads((SAMPLES / "small_polygon.json").read_text())

        region = fake_region("region_id", points)
        # Convert list of tuples to list of list
        result = page_xml_parser.simplify_polygon(region)
        result = [list(ele) for ele in result]

        self.assertListEqual(result, expected)
        self.assertLessEqual(len(expected), 512)

    def test_simplify_polygon_out_of_bounds(self):
        page_xml_parser = PageXmlParser(EMPTY_XML, smaller_image)
        # out of bounds polygon
        points = json.loads((SAMPLES / "big_polygon.json").read_text())
        region = fake_region("region_id", points)
        result = page_xml_parser.simplify_polygon(region)
        self.assertEqual(result, None)

    def test_simplify_polygon_crop(self):
        page_xml_parser = PageXmlParser(EMPTY_XML, smaller_image)
        points = [
            (200, 10),
            (250, 15),
            (290, 14),
            (330, 12),
            (320, 60),
            (296, 62),
            (245, 64),
            (222, 62),
            (215, 63),
            (200, 10),
        ]
        expected = [
            (200, 10),
            (250, 15),
            (290, 14),
            (300, 12),
            (300, 60),
            (296, 62),
            (245, 64),
            (222, 62),
            (215, 63),
            (200, 10),
        ]
        region = fake_region("region_id", points)
        result = page_xml_parser.simplify_polygon(region)
        self.assertListEqual(result, expected)

    def test_simplify_polygon_negative_points(self):
        page_xml_parser = PageXmlParser(EMPTY_XML, smaller_image)
        points = [(-12, 30), (0, 32), (-9, 120), (50, 16), (4, -1), (-12, 30)]
        expected = [(0, 30), (0, 32), (0, 120), (50, 16), (4, 0), (0, 30)]
        region = fake_region("region_id", points)
        result = page_xml_parser.simplify_polygon(region)
        self.assertEqual(result, expected)

    def test_simplify_polygon_all_negative_points(self):
        page_xml_parser = PageXmlParser(EMPTY_XML, smaller_image)
        points = [(-12, -20), (-78, -56), (-9, -12), (-67, -18)]
        region = fake_region("region_id", points)
        result = page_xml_parser.simplify_polygon(region)
        self.assertEqual(result, None)

    def test_simplify_polygon_error(self):
        page_xml_parser = PageXmlParser(EMPTY_XML, fake_image)

        # No points
        region = fake_region("region_id", None)
        self.assertIsNone(page_xml_parser.simplify_polygon(region))

        # Not enough points
        region = fake_region("region_id", [])
        self.assertIsNone(page_xml_parser.simplify_polygon(region))
        region = fake_region("region_id", [[0, 0]])
        self.assertIsNone(page_xml_parser.simplify_polygon(region))
        region = fake_region("region_id", [(0, 0), (0, 0)])
        self.assertIsNone(page_xml_parser.simplify_polygon(region))

        # Cannot simplify this polygon
        points = []
        for i in range(600):
            points.append((i, i))
        region = fake_region("region_id", points)
        self.assertIsNone(page_xml_parser.simplify_polygon(region))

    def test_check_transcription(self):
        page_xml_parser = PageXmlParser(EMPTY_XML, fake_image)

        # No text
        region = fake_region("region_id", [], None)
        self.assertFalse(page_xml_parser.check_transcription(region))

        # Empty text
        region = fake_region("region_id", [], "    ")
        self.assertFalse(page_xml_parser.check_transcription(region))

        # Ok
        region = fake_region("region_id", [], "my text")
        self.assertTrue(page_xml_parser.check_transcription(region))

    @requests_mock.Mocker()
    @patch(
        "arkindex_tasks.import_files.pagexml.WORKER_RUN_ID",
        "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
    )
    def test_save(self, mock):
        xml_file = (SAMPLES / "transcript.xml").read_bytes()
        page_xml_parser = PageXmlParser(etree.fromstring(xml_file), fake_image)

        # Mock API
        mock.post(
            "/api/v1/elements/create/",
            status_code=201,
            json={
                "id": "element_id",
                "name": "element_name",
                "corpus": {"id": "corpus_id"},
            },
        )
        mock.post("/api/v1/element/element_id/metadata/")
        mock.post(
            "https://arkindex.teklia.com/api/v1/element/element_id/children/bulk/",
            status_code=201,
            json=[{"id": "region1"}, {"id": "region2"}],
        )
        mock.post(
            "https://arkindex.teklia.com/api/v1/element/region1/children/bulk/",
            status_code=201,
            json=[{"id": "line1"}, {"id": "line2"}, {"id": "line3"}, {"id": "line4"}],
        )
        mock.post("/api/v1/element/line1/metadata/")
        mock.post("/api/v1/element/line2/metadata/")
        mock.post("/api/v1/element/line3/metadata/")
        mock.post("/api/v1/element/line4/metadata/")
        mock.post(
            "https://arkindex.teklia.com/api/v1/element/region2/children/bulk/",
            status_code=201,
            json=[
                {"id": "line1"},
                {"id": "line2"},
            ],
        )
        mock.post(
            "https://arkindex.teklia.com/api/v1/transcription/bulk/",
            status_code=201,
            json={
                "transcriptions": [
                    {"id": "transcription1"},
                    {"id": "transcription2"},
                    {"id": "transcription3"},
                    {"id": "transcription4"},
                    {"id": "transcription5"},
                    {"id": "transcription6"},
                    {"id": "transcription7"},
                    {"id": "transcription8"},
                ]
            },
        )

        element, transcriptions = page_xml_parser.save(
            "corpus_id",
            "folder_id",
            "image_id",
            {"slug": "page_type"},
            {"slug": "paragraph_type"},
            {"slug": "line_type"},
        )

        self.assertDictEqual(
            element,
            {"id": "element_id", "name": "element_name", "corpus": {"id": "corpus_id"}},
        )

        self.assertListEqual(
            transcriptions,
            [
                {"index_region": 0, "transcription_id": "transcription1"},
                {"index_region": 1, "transcription_id": "transcription6"},
            ],
        )

    def test_save_no_worker_version(self):
        page_xml_parser = PageXmlParser(EMPTY_XML, fake_image)

        element, transcriptions = page_xml_parser.save(
            "corpus_id",
            "folder_id",
            "image_id",
            {"slug": "page_type"},
            {"slug": "paragraph_type"},
            {"slug": "line_type"},
        )
        self.assertIsNone(element)
        self.assertListEqual(transcriptions, [])

    def test_save_empty_xml(self):
        page_xml_parser = PageXmlParser(EMPTY_XML, fake_image)

        element, transcriptions = page_xml_parser.save(
            "corpus_id",
            "folder_id",
            "image_id",
            {"slug": "page_type"},
            {"slug": "paragraph_type"},
            {"slug": "line_type"},
        )
        self.assertIsNone(element)
        self.assertListEqual(transcriptions, [])

    @requests_mock.Mocker()
    @patch(
        "arkindex_tasks.import_files.pagexml.WORKER_RUN_ID",
        "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
    )
    def test_save_no_lines(self, mock):
        xml_file = (SAMPLES / "no_lines.xml").read_bytes()
        page_xml_parser = PageXmlParser(etree.fromstring(xml_file), fake_image)

        # Mock API
        mock.post(
            "/api/v1/elements/create/",
            status_code=201,
            json={
                "id": "element_id",
                "name": "element_name",
                "corpus": {"id": "corpus_id"},
            },
        )
        mock.post("/api/v1/element/element_id/metadata/")
        mock.post(
            "https://arkindex.teklia.com/api/v1/element/element_id/children/bulk/",
            status_code=201,
            json=[{"id": "region1"}, {"id": "region2"}],
        )
        mock.post(
            "https://arkindex.teklia.com/api/v1/transcription/bulk/",
            status_code=201,
            json={
                "transcriptions": [
                    {"id": "transcription1"},
                    {"id": "transcription2"},
                ]
            },
        )

        element, transcriptions = page_xml_parser.save(
            "corpus_id",
            "folder_id",
            "image_id",
            {"slug": "page_type"},
            {"slug": "paragraph_type"},
            {"slug": "line_type"},
        )

        self.assertDictEqual(
            element,
            {"id": "element_id", "name": "element_name", "corpus": {"id": "corpus_id"}},
        )

        self.assertListEqual(
            transcriptions,
            [
                {"index_region": 0, "transcription_id": "transcription1"},
                {"index_region": 1, "transcription_id": "transcription2"},
            ],
        )

    @requests_mock.Mocker()
    @patch(
        "arkindex_tasks.import_files.pagexml.WORKER_RUN_ID",
        "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
    )
    def test_save_no_valid_regions(self, mock):
        bad_image = {
            "id": "fake_image_id",
            "path": "/path/path/",
            "width": 0,
            "height": 0,
            "url": "http://example.com",
        }

        xml_file = (SAMPLES / "transcript.xml").read_bytes()
        page_xml_parser = PageXmlParser(etree.fromstring(xml_file), bad_image)

        # Mock API
        mock.post(
            "/api/v1/elements/create/",
            status_code=201,
            json={
                "id": "element_id",
                "name": "element_name",
                "corpus": {"id": "corpus_id"},
            },
        )
        mock.post("/api/v1/element/element_id/metadata/")

        element, transcriptions = page_xml_parser.save(
            "corpus_id",
            "folder_id",
            "image_id",
            {"slug": "page_type"},
            {"slug": "paragraph_type"},
            {"slug": "line_type"},
        )

        self.assertDictEqual(
            element,
            {"id": "element_id", "name": "element_name", "corpus": {"id": "corpus_id"}},
        )

        self.assertListEqual(
            transcriptions,
            [],
        )
