# -*- coding: utf-8 -*-
import json
import shutil
import tempfile
from pathlib import Path
from unittest import TestCase
from unittest.mock import MagicMock, call, patch

from arkindex_tasks.import_files.pdf import (
    build_transcription,
    extract_pdf_images,
    extract_pdf_text,
)

SAMPLES = Path(__file__).absolute().parent / "samples"


class TestPdf(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.datafile = {
            "id": "datafileid",
            "name": "file.pdf",
            "content_type": "application/pdf",
            "local_path": SAMPLES / "file.pdf",
        }
        cls.arabic_datafile = {
            "id": "arabicid",
            "name": "arabic.pdf",
            "content_type": "application/pdf",
            "local_path": SAMPLES / "arabic_sample.pdf",
        }

    def setUp(self):
        self.temp_dir = Path(tempfile.mkdtemp())
        pdf_path = self.temp_dir / "file.pdf"
        shutil.copy(str(SAMPLES / "file.pdf"), str(pdf_path))
        self.datafile["local_path"] = pdf_path

    def tearDown(self):
        shutil.rmtree(str(self.temp_dir))

    @patch("arkindex_tasks.import_files.pdf.convert_from_path")
    def test_extract_pdf_images_not_pdf(self, conv_mock):
        bad_datafile = self.datafile.copy()
        bad_datafile["content_type"] = "text/plain"

        with self.assertRaises(ValueError, msg="File file.pdf is not a PDF"):
            extract_pdf_images(bad_datafile)

        self.assertFalse(conv_mock.called)

    @patch("arkindex_tasks.import_files.pdf.distutils.spawn.find_executable")
    def test_extract_pdf_images_missing_executable(self, find_exec_mock):
        find_exec_mock.return_value = False

        with self.assertRaises(AssertionError, msg="Missing pdftoppm in PATH"):
            extract_pdf_images(self.datafile)

        self.assertEqual(find_exec_mock.call_count, 1)
        self.assertEqual(find_exec_mock.call_args, call("pdftoppm"))

    def test_extract_pdf_images(self):
        dest_dir = Path(tempfile.mkdtemp())
        result = extract_pdf_images(self.datafile, dest_dir)
        self.assertListEqual(
            result, [str(dest_dir / "pdf-0001-1.jpg"), str(dest_dir / "pdf-0001-2.jpg")]
        )
        shutil.rmtree(str(dest_dir))

    def test_extract_pdf_text(self):
        ark_pages = [
            {
                "id": "page1",
                "type": "page",
                "zone": {
                    "image": {
                        "width": 1000,
                        "height": 2000,
                    }
                },
            },
            {
                "id": "page2",
                "type": "page",
                "zone": {
                    "image": {
                        "width": 800,
                        "height": 1900,
                    }
                },
            },
        ]

        result = extract_pdf_text(self.datafile["local_path"], ark_pages)

        self.assertTrue("page1" in result)
        self.assertEqual(len(result["page1"]), 10)
        self.assertTrue("page2" in result)
        self.assertEqual(len(result["page2"]), 6)

        check = SAMPLES / "file_text.json"
        self.assertDictEqual(
            json.loads(check.read_text()),
            result,
        )

    def test_extract_pdf_text_orientation(self):
        ark_page = [
            {
                "id": "page1",
                "type": "page",
                "zone": {
                    "image": {
                        "width": 1000,
                        "height": 2000,
                    }
                },
            }
        ]

        result = extract_pdf_text(self.arabic_datafile["local_path"], ark_page)

        self.assertTrue("page1" in result)
        self.assertEqual(len(result["page1"]), 170)

        check = SAMPLES / "arabic_sample.json"
        self.assertDictEqual(
            json.loads(check.read_text()),
            result,
        )

    def test_build_transcription_wrong_polygon(self):
        pdf_element = MagicMock()
        pdf_element.x0 = 0
        pdf_element.x1 = 1
        pdf_element.y0 = 1
        pdf_element.y1 = 1
        pdf_element.get_text.return_value = "hello"

        pdf_page = MagicMock()
        pdf_page.width = 500
        pdf_page.height = 1000

        ark_page = {
            "id": "page1",
            "type": "page",
            "zone": {
                "image": {
                    "width": 1000,
                    "height": 2000,
                }
            },
        }

        # Checking that we don't have None because the text could be empty
        self.assertTrue(pdf_element.get_text().strip())
        result = build_transcription(pdf_element, pdf_page, ark_page)
        self.assertIsNone(result)

    def test_build_transcription_polygon_out_of_bounds(self):
        pdf_element = MagicMock()
        pdf_element.x0 = -50
        pdf_element.x1 = 600
        pdf_element.y0 = -9999
        pdf_element.y1 = 32135
        pdf_element.get_text.return_value = "hello"

        pdf_page = MagicMock()
        pdf_page.width = 500
        pdf_page.height = 1000

        ark_page = {
            "id": "page1",
            "type": "page",
            "zone": {
                "image": {
                    "width": 1000,
                    "height": 2000,
                }
            },
        }

        # Checking that we don't have None because the text could be empty
        self.assertTrue(pdf_element.get_text().strip())
        result = build_transcription(pdf_element, pdf_page, ark_page)
        self.assertDictEqual(
            result,
            {
                "text": "hello",
                "polygon": [
                    [0, 2000],
                    [1000, 2000],
                    [1000, 0],
                    [0, 0],
                ],
                "confidence": 1.0,
                "orientation": "horizontal-lr",
            },
        )
