# -*- coding: utf-8 -*-
import tempfile
from pathlib import Path
from unittest import TestCase
from unittest.mock import patch
from uuid import uuid4
from zipfile import ZipFile

import requests_mock
from apistar.exceptions import ErrorResponse

from arkindex_tasks.base import get_working_dir
from arkindex_tasks.import_files.transkribus import TranskribusElement

SAMPLES = Path(__file__).absolute().parent.parent.parent / "samples"


@patch(
    "arkindex_tasks.import_files.transkribus.retried_request.retry.wait",
    0,
)
class TestTranskribusElement(TestCase):
    @classmethod
    def setUpClass(cls):
        _, cls.archive_path = tempfile.mkstemp()
        cls.archive = ZipFile(cls.archive_path, mode="w")
        cls.archive.write(str(SAMPLES / "img1.jpg"), arcname="0001_123456.jpg")
        cls.archive.write(str(SAMPLES / "transcript.xml"), arcname="0001_123456.xml")

        cls.collection_path = get_working_dir() / str(1234)
        cls.collection_path.mkdir(exist_ok=True)

    @classmethod
    def tearDownClass(cls):
        cls.archive.close()
        Path(cls.archive_path).unlink()
        cls.collection_path.rmdir()

    def setUp(self):
        self.maxDiff = None
        self.element = TranskribusElement(
            self.archive,
            image_filename="0001_123456.jpg",
            xml_filename="0001_123456.xml",
            corpus_id="corpusid",
            folder_id="folderid",
            element_type={"slug": "page", "display_name": "a page"},
            paragraph_type={"slug": "paragraph", "display_name": "a paragraph"},
            line_type={"slug": "line", "display_name": "a line"},
        )

    @requests_mock.Mocker()
    def test_upload_image(self, mock):
        mock.post(
            "https://arkindex.teklia.com/api/v1/image/",
            json={"id": "imageid", "s3_put_url": "http://s3/put"},
        )
        mock.put("http://s3/put", status_code=200)
        updated_image = {
            "id": "imageid",
            "s3_put_url": None,
            "width": 999,
            "height": 999,
            "status": "checked",
        }
        mock.patch(
            "https://arkindex.teklia.com/api/v1/image/imageid/", json=updated_image
        )

        self.element.upload_image()

        self.assertListEqual(
            [(req.method, req.url) for req in mock.request_history],
            [
                ("POST", "https://arkindex.teklia.com/api/v1/image/"),
                ("PUT", "http://s3/put"),
                ("PATCH", "https://arkindex.teklia.com/api/v1/image/imageid/"),
            ],
        )
        post_request, patch_request = mock.request_history[0], mock.request_history[2]
        self.assertDictEqual(
            post_request.json(), {"hash": "b205bcf449188a123efa79b19c06662f"}
        )
        self.assertDictEqual(patch_request.json(), {"status": "checked"})

        self.assertEqual(self.element.image, updated_image)

    @requests_mock.Mocker()
    def test_upload_image_exists(self, mock):
        mock.post(
            "https://arkindex.teklia.com/api/v1/image/",
            status_code=400,
            json={"id": "imageid"},
        )
        mock.get(
            "https://arkindex.teklia.com/api/v1/image/imageid/", json={"id": "imageid"}
        )
        updated_image = {
            "id": "imageid",
            "s3_put_url": None,
            "width": 999,
            "height": 999,
            "status": "checked",
        }
        mock.patch(
            "https://arkindex.teklia.com/api/v1/image/imageid/", json=updated_image
        )

        self.element.upload_image()

        self.assertListEqual(
            [(req.method, req.url) for req in mock.request_history],
            [
                ("POST", "https://arkindex.teklia.com/api/v1/image/"),
                ("GET", "https://arkindex.teklia.com/api/v1/image/imageid/"),
                ("PATCH", "https://arkindex.teklia.com/api/v1/image/imageid/"),
            ],
        )
        post_request, patch_request = mock.request_history[0], mock.request_history[2]
        self.assertDictEqual(
            post_request.json(), {"hash": "b205bcf449188a123efa79b19c06662f"}
        )
        self.assertDictEqual(patch_request.json(), {"status": "checked"})

        self.assertEqual(self.element.image, updated_image)

    @requests_mock.Mocker()
    def test_upload_image_failure(self, mock):
        mock.post(
            "https://arkindex.teklia.com/api/v1/image/",
            status_code=500,
        )

        with self.assertRaises(ErrorResponse) as e:
            self.element.upload_image()

        # Process does not complete, image is not uploaded
        self.assertEqual(e.exception.status_code, 500)
        self.assertListEqual(
            [(req.method, req.url) for req in mock.request_history],
            [
                ("POST", "https://arkindex.teklia.com/api/v1/image/"),
                ("POST", "https://arkindex.teklia.com/api/v1/image/"),
                ("POST", "https://arkindex.teklia.com/api/v1/image/"),
                ("POST", "https://arkindex.teklia.com/api/v1/image/"),
                ("POST", "https://arkindex.teklia.com/api/v1/image/"),
            ],
        )

    @patch(
        "arkindex_tasks.import_files.pagexml.WORKER_RUN_ID",
        "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
    )
    @patch(
        "arkindex_tasks.import_files.transkribus.WORKER_RUN_ID",
        "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
    )
    @requests_mock.Mocker()
    def test_upload_transcriptions(self, mock):
        mock.post(
            "https://arkindex.teklia.com/api/v1/elements/create/",
            json={"id": "elementid", "corpus": {"id": "corpusid"}},
        )
        mock.post("/api/v1/element/elementid/metadata/")
        mock.post(
            "https://arkindex.teklia.com/api/v1/element/elementid/children/bulk/",
            status_code=201,
            json=[{"id": "region1"}, {"id": "region2"}],
        )
        mock.post(
            "https://arkindex.teklia.com/api/v1/element/region1/children/bulk/",
            status_code=201,
            json=[{"id": "line1"}, {"id": "line2"}, {"id": "line3"}, {"id": "line4"}],
        )
        mock.post("/api/v1/element/line1/metadata/")
        mock.post("/api/v1/element/line2/metadata/")
        mock.post("/api/v1/element/line3/metadata/")
        mock.post("/api/v1/element/line4/metadata/")
        mock.post(
            "https://arkindex.teklia.com/api/v1/element/region2/children/bulk/",
            status_code=201,
            json=[
                {"id": "line1"},
                {"id": "line2"},
            ],
        )
        mock.post(
            "https://arkindex.teklia.com/api/v1/transcription/bulk/",
            status_code=201,
            json={
                "transcriptions": [
                    {"id": "transcription1"},
                    {"id": "transcription2"},
                    {"id": "transcription3"},
                    {"id": "transcription4"},
                    {"id": "transcription5"},
                    {"id": "transcription6"},
                    {"id": "transcription7"},
                    {"id": "transcription8"},
                ]
            },
        )

        self.element.page = {
            "id": "elementid",
            "corpus": {"id": "corpusid"},
            "zone": {"image": {"id": "imageid"}},
        }
        self.element.image = {"id": "imageid", "width": 12558699, "height": 12558699}
        self.element.upload_transcriptions()

        self.assertListEqual(
            [(req.method, req.url) for req in mock.request_history],
            [
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/elements/create/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/elementid/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/elementid/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/elementid/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/elementid/children/bulk/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/region1/children/bulk/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/line1/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/line2/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/line3/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/line4/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/region2/children/bulk/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/line1/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/line2/metadata/",
                ),
                ("POST", "https://arkindex.teklia.com/api/v1/transcription/bulk/"),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/elementid/metadata/",
                ),
            ],
        )
        self.assertDictEqual(
            mock.request_history[0].json(),
            {
                "corpus": "corpusid",
                "parent": "folderid",
                "type": "page",
                "name": "42",
                "image": "imageid",
                "worker_run_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            },
        )
        self.assertListEqual(
            [
                req.json()
                for req in mock.request_history
                if "elementid/metadata" in req.url
            ],
            [
                {
                    "type": "reference",
                    "name": "transkribus_page_id",
                    "value": "1",
                    "worker_run_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
                },
                {
                    "name": "folio",
                    "type": "text",
                    "value": "42",
                    "worker_run_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
                },
                {
                    "type": "text",
                    "name": "transkribus_status",
                    "value": "IN_PROGRESS",
                    "worker_run_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
                },
                {
                    "type": "text",
                    "name": "original_image_name",
                    "value": "01R_CE101S01_1907_005.tif",
                    "worker_run_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
                },
            ],
        )

    @patch("arkindex_tasks.import_files.pagexml.WORKER_RUN_ID", str(uuid4()))
    @requests_mock.Mocker()
    def test_run(self, mock):
        mock.post(
            "https://arkindex.teklia.com/api/v1/image/",
            json={
                "id": "imageid",
                "s3_put_url": "http://s3/put",
                "width": 0,
                "height": 0,
            },
        )
        mock.put("http://s3/put", status_code=200)
        updated_image = {
            "id": "imageid",
            "s3_put_url": None,
            "width": 12558699,
            "height": 12558699,
        }
        mock.patch(
            "https://arkindex.teklia.com/api/v1/image/imageid/", json=updated_image
        )
        mock.post(
            "https://arkindex.teklia.com/api/v1/elements/create/",
            json={
                "id": "elementid",
                "name": "new element",
                "corpus": {"id": "corpusid"},
                "zone": {"image": {"id": "imageid"}},
            },
        )
        mock.post("/api/v1/element/elementid/metadata/")
        mock.post(
            "https://arkindex.teklia.com/api/v1/element/elementid/children/bulk/",
            status_code=201,
            json=[{"id": "region1"}, {"id": "region2"}],
        )
        mock.post(
            "https://arkindex.teklia.com/api/v1/element/region1/children/bulk/",
            status_code=201,
            json=[{"id": "line1"}, {"id": "line2"}, {"id": "line3"}, {"id": "line4"}],
        )
        mock.post("/api/v1/element/line1/metadata/")
        mock.post("/api/v1/element/line2/metadata/")
        mock.post("/api/v1/element/line3/metadata/")
        mock.post("/api/v1/element/line4/metadata/")
        mock.post(
            "https://arkindex.teklia.com/api/v1/element/region2/children/bulk/",
            status_code=201,
            json=[
                {"id": "line1"},
                {"id": "line2"},
            ],
        )
        mock.post(
            "https://arkindex.teklia.com/api/v1/transcription/bulk/",
            status_code=201,
            json={
                "transcriptions": [
                    {"id": "transcription1"},
                    {"id": "transcription2"},
                    {"id": "transcription3"},
                    {"id": "transcription4"},
                    {"id": "transcription5"},
                    {"id": "transcription6"},
                    {"id": "transcription7"},
                    {"id": "transcription8"},
                ]
            },
        )

        self.element.run()

        self.assertListEqual(
            [(req.method, req.url) for req in mock.request_history],
            [
                ("POST", "https://arkindex.teklia.com/api/v1/image/"),
                ("PUT", "http://s3/put"),
                ("PATCH", "https://arkindex.teklia.com/api/v1/image/imageid/"),
                ("POST", "https://arkindex.teklia.com/api/v1/elements/create/"),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/elementid/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/elementid/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/elementid/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/elementid/children/bulk/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/region1/children/bulk/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/line1/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/line2/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/line3/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/line4/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/region2/children/bulk/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/line1/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/line2/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/transcription/bulk/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/elementid/metadata/",
                ),
            ],
        )

        self.assertEqual(self.element.image, updated_image)
