# -*- coding: utf-8 -*-
import tempfile
from pathlib import Path
from unittest import TestCase
from unittest.mock import patch
from uuid import uuid4
from zipfile import ZipFile

import requests_mock
from apistar.exceptions import ErrorResponse

from arkindex_tasks.import_files.transkribus import TranskribusImporter

SAMPLES = Path(__file__).absolute().parent.parent.parent / "samples"
EXPECTED = SAMPLES / "expected"

CORPUS_INFO = {
    "id": "corpusid",
    "name": "Korpus",
    "types": [
        {"slug": "book", "display_name": "a book", "folder": True},
        {"slug": "page", "display_name": "a page", "folder": False},
        {
            "slug": "paragraph",
            "display_name": "a paragraph",
            "folder": False,
            "allowed_transcription": "paragraph",
        },
        {
            "slug": "line",
            "display_name": "a line",
            "folder": False,
            "allowed_transcription": "line",
        },
    ],
}


class TestTranskribusImporter(TestCase):
    @classmethod
    def setUpClass(cls):
        _, cls.archive_path = tempfile.mkstemp()
        cls.archive_path = Path(cls.archive_path)
        with ZipFile(cls.archive_path, mode="w") as archive:
            archive.write(
                str(SAMPLES / "img1.jpg"), arcname="555/document/0001_123456.jpg"
            )
            archive.write(
                str(SAMPLES / "transcript.xml"), arcname="555/document/0001_123456.xml"
            )

        _, cls.subdir_archive_path = tempfile.mkstemp()
        cls.subdir_archive_path = Path(cls.subdir_archive_path)
        with ZipFile(cls.subdir_archive_path, mode="w") as archive:
            archive.write(
                str(SAMPLES / "img1.jpg"),
                arcname="555/document/some/subdir/abcdefg.jpg",
            )
            archive.write(
                str(SAMPLES / "transcript.xml"),
                arcname="555/document/foo/bar/abcdefg.xml",
            )

    @classmethod
    def tearDownClass(cls):
        cls.archive_path.unlink()

    def setUp(self):
        self.maxDiff = None

    @requests_mock.Mocker()
    @patch("arkindex_tasks.import_files.transkribus.WORKER_RUN_ID", str(uuid4()))
    def test_init(self, mock):
        mock.get("https://arkindex.teklia.com/api/v1/corpus/notfound/", status_code=404)
        mock.get(
            "https://arkindex.teklia.com/api/v1/corpus/corpusid/", json=CORPUS_INFO
        )

        with self.assertRaises(ErrorResponse):
            TranskribusImporter(
                self.archive_path,
                corpus="notfound",
                folder_type="book",
                element_type="page",
                paragraph_type="paragraph",
                line_type="line",
            )

        with self.assertRaises(ValueError, msg="Type hook not found in corpus Korpus"):
            TranskribusImporter(
                self.archive_path,
                corpus="corpusid",
                folder_type="hook",
                element_type="page",
                paragraph_type="paragraph",
                line_type="line",
            )

        with self.assertRaises(ValueError, msg="Type mage not found in corpus Korpus"):
            TranskribusImporter(
                self.archive_path,
                corpus="corpusid",
                folder_type="book",
                element_type="mage",
                paragraph_type="paragraph",
                line_type="line",
            )

        with self.assertRaises(ValueError, msg="Type page is not a folder type"):
            TranskribusImporter(
                self.archive_path,
                corpus="corpusid",
                folder_type="page",
                element_type="page",
                paragraph_type="paragraph",
                line_type="line",
            )

        with self.assertRaises(ValueError, msg="Type book should not be a folder type"):
            TranskribusImporter(
                self.archive_path,
                corpus="corpusid",
                folder_type="book",
                element_type="book",
                paragraph_type="paragraph",
                line_type="line",
            )

        with self.assertRaises(
            ValueError, msg="Type paragraf not found in corpus Korpus"
        ):
            TranskribusImporter(
                self.archive_path,
                corpus="corpusid",
                folder_type="book",
                element_type="page",
                paragraph_type="paragraf",
                line_type="line",
            )

        with self.assertRaises(ValueError, msg="Type book should not be a folder type"):
            TranskribusImporter(
                self.archive_path,
                corpus="corpusid",
                folder_type="book",
                element_type="page",
                paragraph_type="book",
                line_type="line",
            )

        with self.assertRaises(ValueError, msg="Type lin not found in corpus Korpus"):
            TranskribusImporter(
                self.archive_path,
                corpus="corpusid",
                folder_type="book",
                element_type="page",
                paragraph_type="paragraph",
                line_type="lin",
            )

        with self.assertRaises(ValueError, msg="Type book should not be a folder type"):
            TranskribusImporter(
                self.archive_path,
                corpus="corpusid",
                folder_type="book",
                element_type="page",
                paragraph_type="paragraph",
                line_type="book",
            )

        importer = TranskribusImporter(
            self.archive_path,
            corpus="corpusid",
            folder_type="book",
            element_type="page",
            paragraph_type="paragraph",
            line_type="line",
        )

        self.assertEqual(importer.archive_path, self.archive_path)
        self.assertDictEqual(importer.corpus, CORPUS_INFO)
        self.assertDictEqual(
            importer.folder_type,
            {"slug": "book", "display_name": "a book", "folder": True},
        )
        self.assertDictEqual(
            importer.element_type,
            {"slug": "page", "display_name": "a page", "folder": False},
        )
        self.assertDictEqual(
            importer.paragraph_type,
            {
                "slug": "paragraph",
                "display_name": "a paragraph",
                "folder": False,
                "allowed_transcription": "paragraph",
            },
        )
        self.assertDictEqual(
            importer.line_type,
            {
                "slug": "line",
                "display_name": "a line",
                "folder": False,
                "allowed_transcription": "line",
            },
        )

    @patch("arkindex_tasks.import_files.pagexml.WORKER_RUN_ID", str(uuid4()))
    @patch("arkindex_tasks.import_files.transkribus.WORKER_RUN_ID", str(uuid4()))
    @requests_mock.Mocker()
    def test_run(self, mock):
        mock.get(
            "https://arkindex.teklia.com/api/v1/corpus/corpusid/", json=CORPUS_INFO
        )
        mock.get(
            "https://arkindex.teklia.com/api/v1/corpus/corpusid/elements/?name=document&type=book",
            json={"count": 0, "number": 1, "results": []},
        )
        mock.post(
            "https://arkindex.teklia.com/api/v1/image/",
            json={
                "id": "imageid",
                "s3_put_url": "http://s3/put",
                "width": 0,
                "height": 0,
            },
        )
        mock.put("http://s3/put", status_code=200)
        mock.patch(
            "https://arkindex.teklia.com/api/v1/image/imageid/",
            json={
                "id": "imageid",
                "s3_put_url": None,
                "width": 12558699,
                "height": 12558699,
            },
        )
        mock.post(
            "https://arkindex.teklia.com/api/v1/elements/create/",
            [
                {"json": {"id": "folderid"}},
                {
                    "json": {
                        "id": "elementid",
                        "name": "new element",
                        "corpus": {"id": "corpusid"},
                        "zone": {"image": {"id": "imageid"}},
                    }
                },
            ],
        )
        mock.post("/api/v1/element/elementid/metadata/")
        mock.post(
            "https://arkindex.teklia.com/api/v1/element/elementid/children/bulk/",
            status_code=201,
            json=[{"id": "region1"}, {"id": "region2"}],
        )
        mock.post(
            "https://arkindex.teklia.com/api/v1/element/region1/children/bulk/",
            status_code=201,
            json=[{"id": "line1"}, {"id": "line2"}, {"id": "line3"}, {"id": "line4"}],
        )
        mock.post("/api/v1/element/line1/metadata/")
        mock.post("/api/v1/element/line2/metadata/")
        mock.post("/api/v1/element/line3/metadata/")
        mock.post("/api/v1/element/line4/metadata/")
        mock.post(
            "https://arkindex.teklia.com/api/v1/element/region2/children/bulk/",
            status_code=201,
            json=[
                {"id": "line1"},
                {"id": "line2"},
            ],
        )
        mock.post(
            "https://arkindex.teklia.com/api/v1/transcription/bulk/",
            status_code=201,
            json={
                "transcriptions": [
                    {"id": "transcription1"},
                    {"id": "transcription2"},
                    {"id": "transcription3"},
                    {"id": "transcription4"},
                    {"id": "transcription5"},
                    {"id": "transcription6"},
                    {"id": "transcription7"},
                    {"id": "transcription8"},
                ]
            },
        )

        importer = TranskribusImporter(
            self.archive_path,
            corpus="corpusid",
            folder_type="book",
            element_type="page",
            paragraph_type="paragraph",
            line_type="line",
        )

        self.assertListEqual(
            importer.run(),
            [
                {"id": "folderid"},
                {
                    "id": "elementid",
                    "name": "new element",
                    "corpus": {"id": "corpusid"},
                    "zone": {"image": {"id": "imageid"}},
                },
            ],
        )

        self.assertListEqual(
            [(req.method, req.url) for req in mock.request_history],
            [
                ("GET", "https://arkindex.teklia.com/api/v1/corpus/corpusid/"),
                (
                    "GET",
                    "https://arkindex.teklia.com/api/v1/corpus/corpusid/elements/?name=document&type=book",
                ),
                ("POST", "https://arkindex.teklia.com/api/v1/elements/create/"),
                ("POST", "https://arkindex.teklia.com/api/v1/image/"),
                ("PUT", "http://s3/put"),
                ("PATCH", "https://arkindex.teklia.com/api/v1/image/imageid/"),
                ("POST", "https://arkindex.teklia.com/api/v1/elements/create/"),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/elementid/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/elementid/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/elementid/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/elementid/children/bulk/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/region1/children/bulk/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/line1/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/line2/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/line3/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/line4/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/region2/children/bulk/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/line1/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/line2/metadata/",
                ),
                ("POST", "https://arkindex.teklia.com/api/v1/transcription/bulk/"),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/elementid/metadata/",
                ),
            ],
        )

    @patch("arkindex_tasks.import_files.pagexml.WORKER_RUN_ID", str(uuid4()))
    @patch("arkindex_tasks.import_files.transkribus.WORKER_RUN_ID", str(uuid4()))
    @requests_mock.Mocker()
    def test_run_subdir(self, mock):
        mock.get(
            "https://arkindex.teklia.com/api/v1/corpus/corpusid/", json=CORPUS_INFO
        )
        mock.get(
            "https://arkindex.teklia.com/api/v1/corpus/corpusid/elements/?name=document&type=book",
            json={"count": 0, "number": 1, "results": []},
        )
        mock.post(
            "https://arkindex.teklia.com/api/v1/image/",
            json={
                "id": "imageid",
                "s3_put_url": "http://s3/put",
                "width": 0,
                "height": 0,
            },
        )
        mock.put("http://s3/put", status_code=200)
        mock.patch(
            "https://arkindex.teklia.com/api/v1/image/imageid/",
            json={
                "id": "imageid",
                "s3_put_url": None,
                "width": 12558699,
                "height": 12558699,
            },
        )
        mock.post(
            "https://arkindex.teklia.com/api/v1/elements/create/",
            [
                {"json": {"id": "folderid"}},
                {
                    "json": {
                        "id": "elementid",
                        "name": "new element",
                        "corpus": {"id": "corpusid"},
                        "zone": {"image": {"id": "imageid"}},
                    }
                },
            ],
        )
        mock.post("/api/v1/element/elementid/metadata/")
        mock.post(
            "https://arkindex.teklia.com/api/v1/element/elementid/children/bulk/",
            status_code=201,
            json=[{"id": "region1"}, {"id": "region2"}],
        )
        mock.post(
            "https://arkindex.teklia.com/api/v1/element/region1/children/bulk/",
            status_code=201,
            json=[{"id": "line1"}, {"id": "line2"}, {"id": "line3"}, {"id": "line4"}],
        )
        mock.post("/api/v1/element/line1/metadata/")
        mock.post("/api/v1/element/line2/metadata/")
        mock.post("/api/v1/element/line3/metadata/")
        mock.post("/api/v1/element/line4/metadata/")
        mock.post(
            "https://arkindex.teklia.com/api/v1/element/region2/children/bulk/",
            status_code=201,
            json=[
                {"id": "line1"},
                {"id": "line2"},
            ],
        )
        mock.post(
            "https://arkindex.teklia.com/api/v1/transcription/bulk/",
            status_code=201,
            json={
                "transcriptions": [
                    {"id": "transcription1"},
                    {"id": "transcription2"},
                    {"id": "transcription3"},
                    {"id": "transcription4"},
                    {"id": "transcription5"},
                    {"id": "transcription6"},
                    {"id": "transcription7"},
                    {"id": "transcription8"},
                ]
            },
        )

        importer = TranskribusImporter(
            self.subdir_archive_path,
            corpus="corpusid",
            folder_type="book",
            element_type="page",
            paragraph_type="paragraph",
            line_type="line",
        )

        self.assertListEqual(
            importer.run(),
            [
                {"id": "folderid"},
                {
                    "id": "elementid",
                    "name": "new element",
                    "corpus": {"id": "corpusid"},
                    "zone": {"image": {"id": "imageid"}},
                },
            ],
        )

        self.assertListEqual(
            [(req.method, req.url) for req in mock.request_history],
            [
                ("GET", "https://arkindex.teklia.com/api/v1/corpus/corpusid/"),
                (
                    "GET",
                    "https://arkindex.teklia.com/api/v1/corpus/corpusid/elements/?name=document&type=book",
                ),
                ("POST", "https://arkindex.teklia.com/api/v1/elements/create/"),
                ("POST", "https://arkindex.teklia.com/api/v1/image/"),
                ("PUT", "http://s3/put"),
                ("PATCH", "https://arkindex.teklia.com/api/v1/image/imageid/"),
                ("POST", "https://arkindex.teklia.com/api/v1/elements/create/"),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/elementid/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/elementid/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/elementid/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/elementid/children/bulk/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/region1/children/bulk/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/line1/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/line2/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/line3/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/line4/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/region2/children/bulk/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/line1/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/line2/metadata/",
                ),
                ("POST", "https://arkindex.teklia.com/api/v1/transcription/bulk/"),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/elementid/metadata/",
                ),
            ],
        )

    @patch("arkindex_tasks.import_files.pagexml.WORKER_RUN_ID", str(uuid4()))
    @patch("arkindex_tasks.import_files.transkribus.WORKER_RUN_ID", str(uuid4()))
    @requests_mock.Mocker()
    def test_run_parent(self, mock):
        mock.get(
            "https://arkindex.teklia.com/api/v1/corpus/corpusid/", json=CORPUS_INFO
        )
        mock.get(
            "https://arkindex.teklia.com/api/v1/element/parentid/",
            json={
                "id": "parentid",
                "corpus": {"id": "corpusid"},
            },
        )
        mock.get(
            "https://arkindex.teklia.com/api/v1/elements/parentid/children/?name=document&type=book&recursive=true",
            json={"count": 0, "number": 1, "results": []},
        )
        mock.post(
            "https://arkindex.teklia.com/api/v1/image/",
            json={
                "id": "imageid",
                "s3_put_url": "http://s3/put",
                "width": 0,
                "height": 0,
            },
        )
        mock.put("http://s3/put", status_code=200)
        mock.patch(
            "https://arkindex.teklia.com/api/v1/image/imageid/",
            json={
                "id": "imageid",
                "s3_put_url": None,
                "width": 12558699,
                "height": 12558699,
            },
        )
        mock.post(
            "https://arkindex.teklia.com/api/v1/elements/create/",
            [
                {"json": {"id": "folderid"}},
                {
                    "json": {
                        "id": "elementid",
                        "name": "new element",
                        "corpus": {"id": "corpusid"},
                        "zone": {"image": {"id": "imageid"}},
                    }
                },
            ],
        )
        mock.post("/api/v1/element/elementid/metadata/")
        mock.post(
            "https://arkindex.teklia.com/api/v1/element/elementid/children/bulk/",
            status_code=201,
            json=[{"id": "region1"}, {"id": "region2"}],
        )
        mock.post(
            "https://arkindex.teklia.com/api/v1/element/region1/children/bulk/",
            status_code=201,
            json=[{"id": "line1"}, {"id": "line2"}, {"id": "line3"}, {"id": "line4"}],
        )
        mock.post("/api/v1/element/line1/metadata/")
        mock.post("/api/v1/element/line2/metadata/")
        mock.post("/api/v1/element/line3/metadata/")
        mock.post("/api/v1/element/line4/metadata/")
        mock.post(
            "https://arkindex.teklia.com/api/v1/element/region2/children/bulk/",
            status_code=201,
            json=[
                {"id": "line1"},
                {"id": "line2"},
            ],
        )
        mock.post(
            "https://arkindex.teklia.com/api/v1/transcription/bulk/",
            status_code=201,
            json={
                "transcriptions": [
                    {"id": "transcription1"},
                    {"id": "transcription2"},
                    {"id": "transcription3"},
                    {"id": "transcription4"},
                    {"id": "transcription5"},
                    {"id": "transcription6"},
                    {"id": "transcription7"},
                    {"id": "transcription8"},
                ]
            },
        )

        importer = TranskribusImporter(
            self.archive_path,
            corpus="corpusid",
            parent_id="parentid",
            folder_type="book",
            element_type="page",
            paragraph_type="paragraph",
            line_type="line",
        )

        self.assertListEqual(
            importer.run(),
            [
                {"id": "folderid"},
                {
                    "id": "elementid",
                    "name": "new element",
                    "corpus": {"id": "corpusid"},
                    "zone": {"image": {"id": "imageid"}},
                },
            ],
        )

        self.assertListEqual(
            [(req.method, req.url) for req in mock.request_history],
            [
                ("GET", "https://arkindex.teklia.com/api/v1/corpus/corpusid/"),
                ("GET", "https://arkindex.teklia.com/api/v1/element/parentid/"),
                (
                    "GET",
                    "https://arkindex.teklia.com/api/v1/elements/parentid/children/?name=document&recursive=True&type=book",
                ),
                ("POST", "https://arkindex.teklia.com/api/v1/elements/create/"),
                ("POST", "https://arkindex.teklia.com/api/v1/image/"),
                ("PUT", "http://s3/put"),
                ("PATCH", "https://arkindex.teklia.com/api/v1/image/imageid/"),
                ("POST", "https://arkindex.teklia.com/api/v1/elements/create/"),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/elementid/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/elementid/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/elementid/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/elementid/children/bulk/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/region1/children/bulk/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/line1/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/line2/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/line3/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/line4/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/region2/children/bulk/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/line1/metadata/",
                ),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/line2/metadata/",
                ),
                ("POST", "https://arkindex.teklia.com/api/v1/transcription/bulk/"),
                (
                    "POST",
                    "https://arkindex.teklia.com/api/v1/element/elementid/metadata/",
                ),
            ],
        )

    @requests_mock.Mocker()
    @patch("arkindex_tasks.import_files.transkribus.WORKER_RUN_ID", str(uuid4()))
    @patch("arkindex_tasks.import_files.transkribus.TranskribusFolder.run")
    def test_run_folder_error(self, req_mock, run_mock):
        req_mock.get(
            "https://arkindex.teklia.com/api/v1/corpus/corpusid/", json=CORPUS_INFO
        )
        run_mock.side_effect = TypeError("Oh snap!")

        importer = TranskribusImporter(
            self.archive_path,
            corpus="corpusid",
            folder_type="book",
            element_type="page",
            paragraph_type="paragraph",
            line_type="line",
        )
        # Nothing is imported
        self.assertListEqual(importer.run(), [])

        self.assertListEqual(
            [(req.method, req.url) for req in req_mock.request_history],
            [
                ("GET", "https://arkindex.teklia.com/api/v1/corpus/corpusid/"),
            ],
        )
        self.assertEqual(run_mock.call_count, 1)

    @requests_mock.Mocker()
    @patch("arkindex_tasks.import_files.transkribus.WORKER_RUN_ID", str(uuid4()))
    def test_run_parent_wrong_corpus(self, req_mock):
        req_mock.get(
            "https://arkindex.teklia.com/api/v1/corpus/corpusid/", json=CORPUS_INFO
        )
        req_mock.get(
            "https://arkindex.teklia.com/api/v1/element/parentid/",
            json={"id": "parentid", "corpus": {"id": "not_corpusid"}},
        )

        with self.assertRaises(ValueError) as ctx:
            TranskribusImporter(
                self.archive_path,
                corpus="corpusid",
                parent_id="parentid",
                folder_type="book",
                element_type="page",
                paragraph_type="paragraph",
                line_type="line",
            )

        self.assertEqual(
            ctx.exception.args, ("Parent element parentid is not in corpus corpusid",)
        )

        self.assertListEqual(
            [(req.method, req.url) for req in req_mock.request_history],
            [
                ("GET", "https://arkindex.teklia.com/api/v1/corpus/corpusid/"),
                ("GET", "https://arkindex.teklia.com/api/v1/element/parentid/"),
            ],
        )
