# -*- coding: utf-8 -*-
import shutil
import tempfile
from pathlib import Path
from unittest import TestCase
from unittest.mock import patch

import requests_mock

import arkindex_tasks.base

SAMPLES = Path(__file__).absolute().parent / "samples"


@patch("teklia_toolbox.requests.download_file.retry.wait.wait_fixed", new=0)
class TestBase(TestCase):
    def setUp(self):
        self.tmp_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(Path(self.tmp_dir))

    def test_default_working_dir(self):
        arkindex_tasks.base._working_dir = None
        try:
            with patch.dict(
                "arkindex_tasks.base.os.environ", {"PONOS_DATA": "/directories/default"}
            ):
                self.assertEqual(
                    arkindex_tasks.base.get_working_dir(),
                    Path("/directories/default/current"),
                )
        finally:
            arkindex_tasks.base._working_dir = None

    def test_chunks(self):
        self.assertListEqual(
            list(arkindex_tasks.base.split_chunks([i for i in range(1, 6)], 2)),
            [[1, 3, 5], [2, 4]],
        )
        self.assertListEqual(
            list(arkindex_tasks.base.split_chunks([1, 2, 3], 4)), [[1], [2], [3], []]
        )

    @requests_mock.Mocker()
    def test_load_process(self, mock):
        mock.get("/api/v1/process/processid/", json={"id": "processid"})
        task = arkindex_tasks.base.ProcessTask("processid")
        self.assertEqual(task.process_id, "processid")
        self.assertIsNone(task.process)
        task.load_process()
        self.assertDictEqual(task.process, {"id": "processid"})

    @requests_mock.Mocker()
    @patch("arkindex_tasks.base.tempfile")
    def test_download_files(self, mock, temp_mock):
        mock.get(
            "/api/v1/process/file/file1/",
            json={"id": "file1", "name": "img1.jpg", "s3_url": "http://s3/img1.jpg"},
        )
        mock.get("http://s3/img1.jpg", body=open(SAMPLES / "img1.jpg", "rb"))

        temp_mock.mkdtemp.return_value = self.tmp_dir

        task = arkindex_tasks.base.ProcessTask("processid")
        task.process = {"id": "processid", "files": ["file1"]}

        result = task.download_files()
        self.assertListEqual(
            result,
            [
                {
                    "id": "file1",
                    "name": "img1.jpg",
                    "s3_url": "http://s3/img1.jpg",
                    "local_path": Path(self.tmp_dir) / "file1",
                }
            ],
        )

        self.assertTrue((Path(self.tmp_dir) / "file1").is_file())
        self.assertEqual(mock.call_count, 2)

    @requests_mock.Mocker()
    def test_download_files_error(self, mock):
        mock.get(
            "/api/v1/process/file/file1/",
            json={"id": "file1", "name": "img1.jpg", "s3_url": "http://nowhere1/"},
        )
        mock.get(
            "/api/v1/process/file/file2/",
            json={"id": "file2", "name": "img2.jpg", "s3_url": "http://nowhere2/"},
        )
        mock.get("http://nowhere1/", status_code=404)
        mock.get("http://nowhere2/", status_code=404)

        task = arkindex_tasks.base.ProcessTask("processid")
        task.process = {"id": "processid", "files": ["file1", "file2"]}

        result = task.download_files()
        self.assertListEqual(result, [])
        self.assertEqual(mock.call_count, 8)
        self.assertListEqual(
            [(req.method, req.url) for req in mock.request_history],
            [
                ("GET", "https://arkindex.teklia.com/api/v1/process/file/file1/"),
                ("GET", "http://nowhere1/"),
                ("GET", "http://nowhere1/"),
                ("GET", "http://nowhere1/"),
                ("GET", "https://arkindex.teklia.com/api/v1/process/file/file2/"),
                ("GET", "http://nowhere2/"),
                ("GET", "http://nowhere2/"),
                ("GET", "http://nowhere2/"),
            ],
        )

    @requests_mock.Mocker()
    def test_delete_files(self, mock):
        mock.delete("/api/v1/process/file/file1/", status_code=404)
        mock.delete("/api/v1/process/file/file2/", status_code=204)

        task = arkindex_tasks.base.ProcessTask("processid")
        task.delete_files(
            [{"id": "file1", "name": "img1.jpg"}, {"id": "file2", "name": "img2.jpg"}]
        )

        self.assertEqual(mock.call_count, 2)
        self.assertListEqual(
            [(req.method, req.url) for req in mock.request_history],
            [
                ("DELETE", "https://arkindex.teklia.com/api/v1/process/file/file1/"),
                ("DELETE", "https://arkindex.teklia.com/api/v1/process/file/file2/"),
            ],
        )
